import uuid
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from timeit import default_timer
from typing import Any, Callable, Dict, List, Optional, Tuple

# Tests declared with the name _, and with the @test decorator
# have to be stored in here, so that they can later be retrieved.
# They cannot be retrieved directly from the module due to name
# clashes. When we're later looking for tests inside the module,
# we can retrieve any anonymous tests from this dict.
# Map of module absolute Path to list of tests in the module
COLLECTED_TESTS: Dict[Path, List[Callable]] = defaultdict(list)


@dataclass
class Each:
    args: Tuple[Any]

    def __getitem__(self, args):
        return self.args[args]

    def __len__(self):
        return len(self.args)


def _generate_id():
    return uuid.uuid4().hex


class _FormatDict(dict):
    def __missing__(self, key):
        return "{" + key + "}"


def is_test_module_name(module_name: str) -> bool:
    return module_name.startswith("test_") or module_name.endswith("_test")


class _Timer:
    def __init__(self, duration: Optional[float] = None):
        self._start_time = None
        self.duration = duration

    def __enter__(self):
        self._start_time = default_timer()
        return self

    def __exit__(self, *args):
        self.duration = default_timer() - self._start_time



import difflib
from typing import Iterator, List

import pprintpp
from rich.console import Console, ConsoleOptions, RenderResult
from rich.text import Text


class Diff:
    """Constructs a Diff object to render diff-highlighted code."""

    def __init__(self, lhs, rhs, width, show_symbols=False) -> None:
        self.width = width
        self.lhs = lhs if isinstance(lhs, str) else pprintpp.pformat(lhs, width=width)
        self.rhs = rhs if isinstance(rhs, str) else pprintpp.pformat(rhs, width=width)
        self.show_symbols = show_symbols

    def raw_unified_diff(self) -> Iterator[str]:
        differ = difflib.Differ()
        lines_lhs = self.lhs.splitlines()
        lines_rhs = self.rhs.splitlines()
        return differ.compare(lines_lhs, lines_rhs)

    def build_symbolic_unified_diff(self) -> RenderResult:
        diff = self.raw_unified_diff()
        output_lines = []
        style = "grey"
        last_style = style
        for line in diff:
            if line.startswith("- "):
                style = "green"
                output_line = f"+ {line[2:]}"
            elif line.startswith("+ "):
                style = "red"
                output_line = f"- {line[2:]}"
            elif line.startswith("? "):
                if last_style == "red":
                    output_line = line[:-1].replace("+", "-")
                elif last_style == "green":
                    output_line = line[:-1].replace("-", "+")
            else:
                output_line = line
                style = "gray"
            output_lines.append(Text(output_line, style=style))
            last_style = style if style != "gray" else last_style
        return output_lines

    def rewrite_line(self, line, line_to_rewrite, prev_marker):
        marker_style_map = {
            "-": {
                " ": "green",
                "-": "white on green",
                "^": "white on green",
            },
            "+": {
                " ": "red",
                "+": "white on red",
                "^": "white on red",
            },
        }
        new_line = Text("")
        current_span = []
        # Differ lines start with a 2 letter code, so skip past that
        prev_char = line[2]
        for idx, char in enumerate(line[2:], start=2):
            if prev_marker in ("+", "-"):
                if char != prev_char:
                    style = marker_style_map.get(prev_marker, {}).get(prev_char, None)
                    if style is not None:
                        new_line.append_text(Text("".join(current_span), style=style))
                    current_span = []
                if idx - 2 < len(line_to_rewrite):
                    current_span.append(line_to_rewrite[idx - 2])
            prev_char = char

        # Lines starting with ? aren't guaranteed to be the same length as the lines before them
        #  so some characters may be left over. Add any leftover characters to the output.
        # subtract 2 for code at start
        remaining_index = idx - 2
        if prev_marker == "+":
            new_line.append_text(Text(line_to_rewrite[remaining_index:], style="red"))
        elif prev_marker == "-":
            new_line.append_text(Text(line_to_rewrite[remaining_index:], style="green"))
        return new_line

    def build_unified_diff(self) -> RenderResult:
        diff = self.raw_unified_diff()
        prev_marker = ""
        output_lines: List[Text] = []
        for line in diff:
            if line.startswith("- "):
                output_lines.append(Text(line[2:], style="green"))
            elif line.startswith("+ "):
                output_lines.append(Text(line[2:], style="red"))
            elif line.startswith("? "):
                line_to_rewrite = output_lines[-1].plain
                output_lines[-1] = self.rewrite_line(line, line_to_rewrite, prev_marker)
            else:
                output_lines.append(Text(line[2:]))
            prev_marker = line[0]
        return output_lines

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        if self.show_symbols:
            result = self.build_symbolic_unified_diff()
        else:
            result = self.build_unified_diff()
        yield from result



import asyncio
import collections
import functools
import inspect
import traceback
from bdb import BdbQuit
from contextlib import ExitStack, closing, redirect_stderr, redirect_stdout
from dataclasses import dataclass, field
from enum import Enum, auto
from io import StringIO
from pathlib import Path
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Union,
)

from ward._errors import FixtureError, ParameterisationError
from ward._fixtures import FixtureCache, ScopeKey, is_fixture
from ward._testing import (
    COLLECTED_TESTS,
    Each,
    _FormatDict,
    _generate_id,
    _Timer,
    is_test_module_name,
)
from ward._utilities import get_absolute_path
from ward.fixtures import Fixture
from ward.models import CollectionMetadata, Marker, Scope, SkipMarker, XfailMarker

__all__ = [
    "test",
    "skip",
    "xfail",
    "each",
    "Test",
    "TestOutcome",
    "TestResult",
    "ParamMeta",
]


@dataclass
class ParamMeta:
    instance_index: int = 0
    group_size: int = 1


def each(*args):
    """
    Used to parameterise tests.

    This will likely be deprecated before Ward 1.0.

    See documentation for examples.
    """
    return Each(args)


def skip(
    func_or_reason: Union[str, Callable] = None,
    *,
    reason: str = None,
    when: Union[bool, Callable] = True,
):
    """
    Decorator which can be used to optionally skip tests.

    Args:
        func_or_reason (object): The wrapped test function to skip.
        reason: The reason the test was skipped. May appear in output.
        when: Predicate function. Will be called immediately before the test is executed.
            If it evaluates to True, the test will be skipped. Otherwise the test will run as normal.
    """
    if func_or_reason is None:
        return functools.partial(skip, reason=reason, when=when)

    if isinstance(func_or_reason, str):
        return functools.partial(skip, reason=func_or_reason, when=when)

    func = func_or_reason
    marker = SkipMarker(reason=reason, when=when)
    if hasattr(func, "ward_meta"):
        func.ward_meta.marker = marker
    else:
        func.ward_meta = CollectionMetadata(marker=marker)

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def xfail(
    func_or_reason: Union[str, Callable] = None,
    *,
    reason: str = None,
    when: Union[bool, Callable] = True,
):
    """
    Decorator that can be used to mark a test as "expected to fail".

    Args:
        func_or_reason: The wrapped test function to mark as an expected failure.
        reason: The reason we expect the test to fail. May appear in output.
        when: Predicate function. Will be called immediately before the test is executed.
            If it evaluates to True, the test will be marked as an expected failure.
            Otherwise the test will run as normal.
    """
    if func_or_reason is None:
        return functools.partial(xfail, reason=reason, when=when)

    if isinstance(func_or_reason, str):
        return functools.partial(xfail, reason=func_or_reason, when=when)

    func = func_or_reason
    marker = XfailMarker(reason=reason, when=when)
    if hasattr(func, "ward_meta"):
        func.ward_meta.marker = marker
    else:
        func.ward_meta = CollectionMetadata(marker=marker)

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


@dataclass
class Test:
    """
    Representation of a test case.

    Attributes:
        fn: The Python function object that contains the test code.
        module_name: The name of the module the test is defined in.
        id: A unique UUID4 used to identify the test.
        marker: Attached by the skip and xfail decorators.
        description: The description of the test. A format string that can contain basic Markdown syntax.
        param_meta: If this is a parameterised test, contains info about the parameterisation.
        capture_output: If True, output will be captured for this test.
        sout: Buffer that fills with captured stdout as the test executes.
        serr: Buffer that fills with captured stderr as the test executes.
        ward_meta: Metadata that was attached to the raw functions collected by Ward's decorators.
        timer: Timing information about the test.
        tags: List of tags associated with the test.
    """

    fn: Callable
    module_name: str
    id: str = field(default_factory=_generate_id)
    marker: Optional[Marker] = None
    description: Optional[str] = None
    param_meta: Optional[ParamMeta] = field(default_factory=ParamMeta)
    capture_output: bool = True
    sout: StringIO = field(default_factory=StringIO)
    serr: StringIO = field(default_factory=StringIO)
    ward_meta: CollectionMetadata = field(default_factory=CollectionMetadata)
    timer: Optional["_Timer"] = None
    tags: List[str] = field(default_factory=list)

    def __hash__(self):
        return hash((self.__class__, self.id))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.id == other.id

    # FIXME:fix linter C901
    def run(self, cache: FixtureCache, dry_run=False) -> "TestResult":  # noqa: C901
        with ExitStack() as stack:
            self.timer = stack.enter_context(_Timer())
            if self.capture_output:
                stack.enter_context(redirect_stdout(self.sout))
                stack.enter_context(redirect_stderr(self.serr))

            if dry_run:
                with closing(self.sout), closing(self.serr):
                    result = TestResult(self, TestOutcome.DRYRUN)
                return result

            if isinstance(self.marker, SkipMarker) and self.marker.active:
                with closing(self.sout), closing(self.serr):
                    result = TestResult(self, TestOutcome.SKIP)
                return result

            try:
                resolved_args = self.resolver.resolve_args(cache)
                self.format_description(resolved_args)
                if self.is_async_test:
                    coro = self.fn(**resolved_args)
                    asyncio.get_event_loop().run_until_complete(coro)
                else:
                    self.fn(**resolved_args)
            except FixtureError as e:
                outcome = TestOutcome.FAIL
                error: Optional[Exception] = e
            except BdbQuit:
                # We don't want to treat the user quitting the debugger
                # as an exception, so we'll ignore BdbQuit. This will
                # also prevent a large pdb-internal stack trace flooding
                # the terminal.
                pass
            except (Exception, SystemExit) as e:
                error = e
                if isinstance(self.marker, XfailMarker) and self.marker.active:
                    outcome = TestOutcome.XFAIL
                else:
                    outcome = TestOutcome.FAIL
            else:
                error = None
                if isinstance(self.marker, XfailMarker) and self.marker.active:
                    outcome = TestOutcome.XPASS
                else:
                    outcome = TestOutcome.PASS

        with closing(self.sout), closing(self.serr):
            if outcome in (TestOutcome.PASS, TestOutcome.SKIP):
                result = TestResult(self, outcome)
            else:
                if isinstance(error, AssertionError):
                    error.error_line = traceback.extract_tb(
                        error.__traceback__, limit=-1
                    )[0].lineno
                result = TestResult(
                    self,
                    outcome,
                    error,
                    captured_stdout=self.sout.getvalue(),
                    captured_stderr=self.serr.getvalue(),
                )

        return result

    def fail_with_error(self, error: Exception) -> "TestResult":
        return TestResult(
            self, outcome=TestOutcome.FAIL, error=error, message=str(error)
        )

    @property
    def name(self) -> str:
        """The name of the Python function representing the test."""
        return self.fn.__name__

    @property
    def path(self) -> Path:
        """The pathlib.Path to the test module."""
        return self.fn.ward_meta.path

    @property
    def qualified_name(self) -> str:
        """{module_name}.{test_function_name}"""
        name = self.name or ""
        return f"{self.module_name}.{name}"

    @property
    def is_async_test(self) -> bool:
        """True if the test is defined with 'async def'."""
        return inspect.iscoroutinefunction(inspect.unwrap(self.fn))

    @property
    def line_number(self) -> int:
        """
        The line number the test is defined on. Corresponds to the line the first decorator wrapping the
            test appears on.
        """
        return inspect.getsourcelines(self.fn)[1]

    @property
    def has_deps(self) -> bool:
        return len(self.deps()) > 0

    @property
    def is_parameterised(self) -> bool:
        """
        `True` if a test is parameterised, `False` otherwise.
            A test is considered parameterised if any of its default arguments
            have a value that is an instance of `Each`.
        """
        default_args = self.resolver.get_default_args()
        return any(isinstance(arg, Each) for arg in default_args.values())

    @property
    def resolver(self):
        return TestArgumentResolver(self, self.param_meta.instance_index)

    def scope_key_from(self, scope: Scope) -> ScopeKey:
        if scope == Scope.Test:
            return self.id
        elif scope == Scope.Module:
            return self.path
        else:
            return Scope.Global

    def get_parameterised_instances(self) -> List["Test"]:
        """
        If the test is parameterised, return a list of `Test` objects representing
        each test generated as a result of the parameterisation.
        If the test is not parameterised, return a list containing only the test itself.
        If the test is parameterised incorrectly, for example the number of
        items don't match across occurrences of `each` in the test signature,
        then a `ParameterisationError` is raised.
        """
        if not self.is_parameterised:
            return [self]

        number_of_instances = self.find_number_of_instances()

        generated_tests = []
        for instance_index in range(number_of_instances):
            generated_test = Test(
                fn=self.fn,
                module_name=self.module_name,
                marker=self.marker,
                description=self.description,
                param_meta=ParamMeta(
                    instance_index=instance_index, group_size=number_of_instances
                ),
                capture_output=self.capture_output,
            )
            generated_tests.append(generated_test)
        return generated_tests

    def find_number_of_instances(self) -> int:
        """
        Returns the number of instances that would be generated for the current
        parameterised test.

        A parameterised test is only valid if every instance of `each` contains
        an equal number of items. If the current test is an invalid parameterisation,
        then a `ParameterisationError` is raised.
        """
        default_args = self.resolver.get_default_args()
        lengths = [len(arg) for _, arg in default_args.items() if isinstance(arg, Each)]
        is_valid = len(set(lengths)) in (0, 1)
        if not is_valid:
            raise ParameterisationError(
                f"The test {self.name}/{self.description} is parameterised incorrectly. "
                f"Please ensure all instances of 'each' in the test signature "
                f"are of equal length."
            )
        if len(lengths) == 0:
            return 1
        else:
            return lengths[0]

    def deps(self) -> Mapping[str, inspect.Parameter]:
        return inspect.signature(self.fn).parameters

    def format_description(self, args: Dict[str, Any]) -> str:
        """
        Applies any necessary string formatting to the description,
        given a dictionary `args` of values that will be injected
        into the test.

        This method will mutate the Test by updating the description.
        Returns the newly updated description.
        """

        format_dict = _FormatDict(**args)
        if not self.description:
            self.description = ""

        try:
            self.description = self.description.format_map(format_dict)
        except ValueError:
            pass

        return self.description


def test(description: str, *args, tags: Optional[List[str]] = None, **kwargs):
    """
    Decorator used to indicate that the function it wraps should be collected by Ward.

    Args:
        description: The description of the test. A format string. Resolve fixtures and default params that are injected
            into the test will also be injected into this description before it gets output in the test report.
            The description can contain basic Markdown syntax (bold, italic, backticks for code, etc.).
        tags: An optional list of strings that will 'tag' the test. Many tests can share the same tag, and these
            tags can be used to group tests in some logical manner (for example: by business domain or test type).
            Tagged tests can be queried using the --tags option.
    """

    def decorator_test(func):
        unwrapped = inspect.unwrap(func)
        module_name: str = unwrapped.__module__
        is_home_module: bool = "." not in module_name
        if is_test_module_name(module_name) and is_home_module:
            force_path: Path = kwargs.get("_force_path")
            if force_path:
                path = force_path.absolute()
            else:
                path = get_absolute_path(unwrapped)

            if hasattr(unwrapped, "ward_meta"):
                unwrapped.ward_meta.description = description
                unwrapped.ward_meta.tags = tags
                unwrapped.ward_meta.path = path
            else:
                unwrapped.ward_meta = CollectionMetadata(
                    description=description,
                    tags=tags,
                    path=path,
                )

            collect_into = kwargs.get("_collect_into", COLLECTED_TESTS)
            collect_into[path].append(unwrapped)

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return wrapper

        return func

    return decorator_test


class TestOutcome(Enum):
    """
    Enumeration representing all possible outcomes of an attempt at running a test.

    Attributes:
        PASS: Represents a passing test outcome - no errors raised, no assertions failed, the test ran to completion.
        FAIL: The test failed in some way - e.g. an assertion failed or an exception was raised.
        SKIP: The test was skipped.
        XFAIL: The test was expected to fail, and it did fail.
        XPASS: The test was expected to fail, however it unexpectedly passed.
        DRYRUN: The test was not executed because the test session was a dry-run.
    """

    PASS = auto()
    FAIL = auto()
    SKIP = auto()
    XFAIL = auto()  # expected fail
    XPASS = auto()  # unexpected pass
    DRYRUN = auto()  # tests arent executed during dryruns

    @property
    def display_char(self):
        display_chars = {
            TestOutcome.PASS: ".",
            TestOutcome.FAIL: "F",
            TestOutcome.SKIP: "-",
            TestOutcome.XPASS: "U",
            TestOutcome.XFAIL: "x",
            TestOutcome.DRYRUN: ".",
        }
        assert len(display_chars) == len(TestOutcome)
        return display_chars[self]

    @property
    def display_name(self):
        display_names = {
            TestOutcome.PASS: "Passes",
            TestOutcome.FAIL: "Failures",
            TestOutcome.SKIP: "Skips",
            TestOutcome.XPASS: "Unexpected Passes",
            TestOutcome.XFAIL: "Expected Failures",
            TestOutcome.DRYRUN: "Dry-runs",
        }
        assert len(display_names) == len(TestOutcome)
        return display_names[self]

    @property
    def will_fail_session(self) -> bool:
        return self in {TestOutcome.FAIL, TestOutcome.XPASS}

    @property
    def wont_fail_session(self) -> bool:
        return not self.will_fail_session


@dataclass
class TestResult:
    """
    Represents the result of a single test, and contains data that may have been generated as
    part of the execution of that test (for example captured stdout and exceptions that were raised).

    Attributes:
        test: The test corresponding to this result.
        outcome: The outcome of the test: did it pass, fail, get skipped, etc.
        error: If an exception was raised during test execution, it is stored here.
        message: An arbitrary message that can be associated with the result. Generally empty.
        captured_stdout: A string containing anything that was written to stdout during the execution of the test.
        captured_stderr: A string containing anything that was written to stderr during the execution of the test.
    """

    test: Test
    outcome: TestOutcome
    error: Optional[Exception] = None
    message: str = ""
    captured_stdout: str = ""
    captured_stderr: str = ""


def fixtures_used_directly_by_tests(
    tests: Iterable["Test"],
) -> Mapping[Fixture, Collection["Test"]]:
    test_to_fixtures = {t: t.resolver.fixtures for t in tests}

    fixture_to_tests = collections.defaultdict(list)
    for test, used_fixtures in test_to_fixtures.items():
        for fix in used_fixtures.values():
            fixture_to_tests[fix].append(test)

    return fixture_to_tests


@dataclass
class TestArgumentResolver:
    test: "Test"
    iteration: int

    def resolve_args(self, cache: FixtureCache) -> Dict[str, Any]:
        """
        Resolve fixtures and return the resultant name -> Fixture dict.
        If the argument is not a fixture, the raw argument will be used.
        Resolved values will be stored in fixture_cache, accessible
        using the fixture cache key (See `Fixture.key`).
        """
        if self.test.capture_output:
            with redirect_stdout(self.test.sout), redirect_stderr(self.test.serr):
                return self._resolve_args(cache)
        else:
            return self._resolve_args(cache)

    def _resolve_args(self, cache: FixtureCache) -> Dict[str, Any]:
        args_for_iteration = self._get_args_for_iteration()
        resolved_args: Dict[str, Any] = {}
        for name, arg in args_for_iteration.items():
            if is_fixture(arg):
                resolved = self._resolve_single_arg(arg, cache)
            else:
                resolved = arg
            resolved_args[name] = resolved
        return self._unpack_resolved(resolved_args)

    def _get_args_for_iteration(self):
        if not self.test.has_deps:
            return {}
        default_args = self.get_default_args()
        args_for_iteration: Dict[str, Any] = {}
        for name, arg in default_args.items():
            # In the case of parameterised testing, grab the arg corresponding
            # to the current iteration of the parameterised group of tests.
            if isinstance(arg, Each):
                arg = arg[self.iteration]
            args_for_iteration[name] = arg
        return args_for_iteration

    @property
    def fixtures(self) -> Dict[str, Fixture]:
        return {
            name: Fixture(arg)
            for name, arg in self._get_args_for_iteration().items()
            if is_fixture(arg)
        }

    def get_default_args(
        self, func: Optional[Union[Callable, Fixture]] = None
    ) -> Dict[str, Any]:
        """
        Returns a mapping of test argument names to values.

        This method does no fixture resolution.

        If a value is a fixture function, then the raw fixture
        function is returned as a value in the dict, *not* the `Fixture` object.
        """
        fn = func or self.test.fn
        meta = getattr(fn, "ward_meta", None)
        signature = inspect.signature(fn)

        # Override the signature if @using is present
        if meta:
            bound_args = getattr(fn.ward_meta, "bound_args", None)
            if bound_args:
                bound_args.apply_defaults()
                return bound_args.arguments

        default_binding = signature.bind_partial()
        default_binding.apply_defaults()
        return default_binding.arguments

    def _resolve_single_arg(
        self, arg: Callable, cache: FixtureCache
    ) -> Union[Any, Fixture]:
        """
        Get the fixture return value

        If the fixture has been cached, return the value from the cache.
        Otherwise, call the fixture function and return the value.
        """

        if not hasattr(arg, "ward_meta"):
            return arg

        fixture = Fixture(arg)
        if cache.contains(
            fixture, fixture.scope, self.test.scope_key_from(fixture.scope)
        ):
            return cache.get(
                fixture.key, fixture.scope, self.test.scope_key_from(fixture.scope)
            )

        children_defaults = self.get_default_args(func=arg)
        children_resolved = {}
        for name, child_fixture in children_defaults.items():
            child_resolved = self._resolve_single_arg(child_fixture, cache)
            children_resolved[name] = child_resolved

        try:
            args_to_inject = self._unpack_resolved(children_resolved)
            if fixture.is_generator_fixture:
                fixture.gen = arg(**args_to_inject)
                fixture.resolved_val = next(fixture.gen)
            elif fixture.is_async_generator_fixture:
                fixture.gen = arg(**args_to_inject)
                awaitable = fixture.gen.__anext__()
                fixture.resolved_val = asyncio.get_event_loop().run_until_complete(
                    awaitable
                )
            elif fixture.is_coroutine_fixture:
                fixture.resolved_val = asyncio.get_event_loop().run_until_complete(
                    arg(**args_to_inject)
                )
            else:
                fixture.resolved_val = arg(**args_to_inject)
        except (Exception, SystemExit) as e:
            raise FixtureError(f"Unable to resolve fixture '{fixture.name}'") from e
        scope_key = self.test.scope_key_from(fixture.scope)
        cache.cache_fixture(fixture, scope_key)
        return fixture

    @staticmethod
    def _unpack_resolved(fixture_dict: Dict[str, Any]) -> Dict[str, Any]:
        resolved_vals = {}
        for (k, arg) in fixture_dict.items():
            if isinstance(arg, Fixture):
                resolved_vals[k] = arg.resolved_val
            else:
                resolved_vals[k] = arg
        return resolved_vals



from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from typing import DefaultDict, Dict, Generator, List

from ward._errors import ParameterisationError
from ward._fixtures import FixtureCache
from ward.models import Scope
from ward.testing import Test, TestResult


@dataclass
class Suite:
    tests: List[Test]
    cache: FixtureCache = field(default_factory=FixtureCache)

    @property
    def num_tests(self) -> int:
        """
        Returns: The number of tests in the suite, *before* taking parameterisation into account.
        """
        return len(self.tests)

    @property
    def num_tests_with_parameterisation(self) -> int:
        """
        Returns: The number of tests in the suite, *after* taking parameterisation into account.
        """
        return sum(test.find_number_of_instances() for test in self.tests)

    def _test_counts_per_module(self) -> Dict[Path, int]:
        """
        Returns: A dictionary mapping a module Path to the number of tests that can be found within that module.
        """
        module_paths = [test.path for test in self.tests]
        counts: DefaultDict[Path, int] = defaultdict(int)
        for path in module_paths:
            counts[path] += 1
        return counts

    def generate_test_runs(
        self, dry_run: bool = False
    ) -> Generator[TestResult, None, None]:
        """
        Run tests

        Returns a generator which yields test results
        """
        num_tests_per_module = self._test_counts_per_module()
        for test in self.tests:
            num_tests_per_module[test.path] -= 1
            try:
                generated_tests = test.get_parameterised_instances()
            except ParameterisationError as e:
                yield test.fail_with_error(e)
                continue
            for generated_test in generated_tests:
                yield generated_test.run(self.cache, dry_run=dry_run)
                self.cache.teardown_fixtures_for_scope(
                    Scope.Test, scope_key=generated_test.id
                )

            if num_tests_per_module[test.path] == 0:
                self.cache.teardown_fixtures_for_scope(
                    Scope.Module, scope_key=test.path
                )

        self.cache.teardown_global_fixtures()



import abc
import inspect
import itertools
import math
import os
import platform
import statistics
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from textwrap import dedent
from typing import (
    Collection,
    Dict,
    Generator,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
)

from rich.console import (
    Console,
    ConsoleOptions,
    RenderableType,
    RenderGroup,
    RenderResult,
)
from rich.highlighter import NullHighlighter
from rich.live import Live
from rich.markdown import Markdown
from rich.padding import Padding
from rich.panel import Panel
from rich.pretty import Pretty
from rich.progress import (
    BarColumn,
    Progress,
    RenderableColumn,
    SpinnerColumn,
    TimeElapsedColumn,
)
from rich.rule import Rule
from rich.syntax import Syntax
from rich.table import Table
from rich.text import Text
from rich.theme import Theme
from rich.traceback import Traceback
from rich.tree import Tree

from ward._diff import Diff
from ward._fixtures import FixtureHierarchyMapping, fixture_parents_and_children
from ward._suite import Suite
from ward._utilities import group_by
from ward._ward_version import __version__
from ward.expect import Comparison, TestFailure
from ward.fixtures import Fixture
from ward.models import ExitCode, Scope
from ward.testing import Test, TestOutcome, TestResult, fixtures_used_directly_by_tests

HORIZONTAL_PAD = (0, 1, 0, 1)

INDENT = " " * 2

theme = Theme(
    {
        "title": "bold",
        "heading": "bold",
        "pass": "#ffffff on #137C39",
        "pass.textonly": "#189F4A",
        "fail": "#ffffff on #BF2D2D",
        "fail.textonly": "#BF2D2D",
        "fail.header": "bold #BF2D2D",
        "skip": "#ffffff on #0E67B3",
        "skip.textonly": "#1381E0",
        "xpass": "#162740 on #F4C041",
        "xpass.textonly": "#F4C041",
        "xfail": "#ffffff on #695CC8",
        "xfail.textonly": "#695CC8",
        "muted": "dim",
        "info": "yellow italic",
        "dryrun": "#ffffff on #162740",
        "rule.line": "#189F4A",
        "fixture.name": "bold #1381E0",
        "fixture.scope.test": "bold #189F4A",
        "fixture.scope.module": "bold #F4C041",
        "fixture.scope.global": "bold #EA913C",
        "usedby": "#9285F6",
    }
)
rich_console = Console(theme=theme, highlighter=NullHighlighter())


def format_test_id(test_result: TestResult) -> str:
    """
    Format module name, line number, and test case number
    """
    return f"{format_test_location(test_result.test)}{format_test_case_number(test_result.test)}"


def format_test_location(test: Test) -> str:
    """
    Returns the location of a test as a string of the form '{test.module_name}:{test.line_number}'
    """
    return f"{test.module_name}:{test.line_number}"


def format_test_case_number(test: Test) -> str:
    """
    Returns a string of the format '[{current_test_number}/{num_parameterised_instances}]'.

    For example, for the 3rd run of a test that is parameterised with 5 parameter sets the
    return value is '[3/5]'.
    """
    param_meta = test.param_meta
    if param_meta.group_size > 1:
        pad = len(str(param_meta.group_size))
        iter_indicator = (
            f"[{param_meta.instance_index + 1:>{pad}}/{param_meta.group_size}]"
        )
    else:
        iter_indicator = ""

    return iter_indicator


class TestOutputStyle(str, Enum):
    TEST_PER_LINE = "test-per-line"
    DOTS_GLOBAL = "dots-global"
    DOTS_MODULE = "dots-module"
    LIVE = "live"
    NONE = "none"


class TestProgressStyle(str, Enum):
    INLINE = "inline"
    BAR = "bar"
    NONE = "none"


def get_test_result_line(
    test_result: TestResult,
    test_index: int,
    num_tests: int,
    progress_styles: List[TestProgressStyle],
    extra_left_pad: int = 0,
) -> Table:
    """
    Outputs a single test result to the terminal in Ward's standard output
    format which outputs a single test per line.
    """
    outcome_tag = test_result.outcome.name[:4]

    test = test_result.test
    test_location = format_test_location(test)
    test_case_number = format_test_case_number(test)
    test_style = outcome_to_style(test_result.outcome)

    grid = Table.grid(expand=True)
    grid.add_column()
    grid.add_column()
    grid.add_column()
    columns = [
        Padding(outcome_tag, style=test_style, pad=(0, 1, 0, 1 + extra_left_pad)),
        Padding(f"{test_location}{test_case_number}", style="muted", pad=(0, 1, 0, 1)),
        Padding(
            Markdown(test.description, inline_code_theme="ansi_dark"), pad=(0, 1, 0, 0)
        ),
    ]

    # Skip/Xfail tests may have a reason note attached that we'll print
    reason = getattr(test.marker, "reason", "")
    if reason and test.marker.active:
        grid.add_column(justify="center", style=test_style)
        columns.append(Padding(reason, pad=(0, 1, 0, 1)))

    if TestProgressStyle.INLINE in progress_styles:
        grid.add_column(justify="right", style="muted")
        columns.append(f"{(test_index + 1) / num_tests:>4.0%}")

    grid.add_row(*columns)

    return grid


INLINE_PROGRESS_LEN = 5  # e.g. "  93%"


def get_dot(result: TestResult) -> Text:
    style = outcome_to_style(result.outcome)
    return Text(result.outcome.display_char, style=style, end="")


@dataclass
class TestTimingStatsPanel:
    all_tests_in_session: List[TestResult]
    num_tests_to_show: int

    @property
    def _raw_test_durations_secs(self):
        return [r.test.timer.duration for r in self.all_tests_in_session]

    @property
    def _median_secs(self):
        return statistics.median(self._raw_test_durations_secs)

    @property
    def _percentile99_secs(self):
        data = self._raw_test_durations_secs
        size = len(data)
        percentile = 99
        return sorted(data)[int(math.ceil((size * percentile) / 100)) - 1]

    def __rich_console__(self, c: Console, co: ConsoleOptions) -> RenderResult:
        test_results = sorted(
            self.all_tests_in_session, key=lambda r: r.test.timer.duration, reverse=True
        )
        grid = Table.grid(padding=(0, 2, 0, 0))
        grid.add_column(justify="right")  # Time taken
        grid.add_column()  # Test ID
        grid.add_column()  # Test description

        for result in test_results[: self.num_tests_to_show]:
            time_taken_secs = result.test.timer.duration
            time_taken_millis = time_taken_secs * 1000
            test_id = format_test_id(result)
            description = result.test.description
            grid.add_row(
                f"[b]{time_taken_millis:.0f}[/b]ms",
                Text(test_id, style="muted"),
                description,
            )

        num_slowest_displayed = min(
            len(self.all_tests_in_session), self.num_tests_to_show
        )
        panel = Panel(
            RenderGroup(
                Padding(
                    f"Median: [b]{self._median_secs * 1000:.2f}[/b]ms"
                    f" [muted]|[/muted] "
                    f"99th Percentile: [b]{self._percentile99_secs * 1000:.2f}[/b]ms",
                    pad=(0, 0, 1, 0),
                ),
                grid,
            ),
            title=f"[b white]{num_slowest_displayed} Slowest Tests[/b white]",
            style="none",
            border_style="rule.line",
        )

        yield panel


@dataclass
class SessionPrelude:
    time_to_collect_secs: float
    num_tests_collected: int
    num_fixtures_collected: int
    config_path: Optional[Path]
    python_impl: str = field(default=platform.python_implementation())
    python_version: str = field(default=platform.python_version())
    ward_version: str = field(default=__version__)

    def __rich_console__(self, c: Console, co: ConsoleOptions) -> RenderResult:
        yield Rule(
            Text(
                f"Ward {self.ward_version} | {self.python_impl} {self.python_version}",
                style="title",
            )
        )
        if self.config_path:
            try:
                path = self.config_path.relative_to(Path.cwd())
            except ValueError:
                path = self.config_path.name
            yield f"Loaded config from [b]{path}[/b]."

        yield (
            f"Found [b]{self.num_tests_collected}[/b] tests "
            f"and [b]{self.num_fixtures_collected}[/b] fixtures "
            f"in [b]{self.time_to_collect_secs:.2f}[/b] seconds."
        )


class ResultProcessor(abc.ABC):
    @abc.abstractmethod
    def handle_result(self, test_result: TestResult):
        pass


class TerminalResultProcessor(ResultProcessor):
    def __init__(
        self,
        suite: Suite,
        test_output_style: str,
        progress_styles: List[TestProgressStyle],
        config_path: Optional[Path],
        show_diff_symbols: bool = False,
    ):
        self.suite = suite
        self.test_output_style = test_output_style
        self.progress_styles = progress_styles
        self.config_path = config_path
        self.show_diff_symbols = show_diff_symbols

    def handle_result(self, test_result: TestResult):
        # Make the actual output of the result a pluggy hook, so that users can implement their own version
        pass


class TestResultDisplayWidget:
    def __init__(self, num_tests: int, progress_styles: List[TestProgressStyle]):
        self.console = rich_console
        self.num_tests = num_tests
        self.progress_styles = progress_styles

    def footer(self, test_results: List[TestResult]) -> Optional[RenderableType]:
        """
        This method should return an object that can be rendered by Rich.
        It will be inserted into the "footer" of the test suite result display,
        which hugs the bottom of the output as the suite runs.

        This method may be called at any time to refresh the state of the footer,
        so it should be a pure function.

        If this function returns ``None``, it will not cause anything to be
        rendered in the footer. You can use this to "hide" the footer based
        on state captured during the suite.
        """
        pass

    def after_test(self, test_index: int, test_result: TestResult) -> None:
        """
        This method is called after each test is executed,
        with the results of that test and the index of that test in the suite.

        Some ways you can use this method:
         - Capture state for use in other methods of your widget.
         - Print to the terminal using the attached Console (``self.console``).
           Anything printed this way will appear above the footer
           and will persist after the suite is done.
        """
        pass

    def after_suite(self, test_results: List[TestResult]) -> None:
        """
        This method is called after the suite is done executing
        (or is cancelled, or aborts mid-run, etc.),
        with results for all of the tests that have been run.

        Some ways you can use this method:
         - Change the return value of your footer to None to prevent it
           from appearing in the final persistent output.
        """
        pass


class TestPerLine(TestResultDisplayWidget):
    def after_test(self, test_index: int, test_result: TestResult) -> None:
        self.console.print(
            get_test_result_line(
                test_result, test_index, self.num_tests, self.progress_styles
            )
        )


class DotsDisplayWidget(TestResultDisplayWidget, abc.ABC):
    def __init__(self, num_tests: int, progress_styles: List[TestProgressStyle]):
        super().__init__(num_tests, progress_styles)

        self.base_max_dots_per_line = get_terminal_size().width
        if TestProgressStyle.INLINE in progress_styles:
            self.base_max_dots_per_line -= INLINE_PROGRESS_LEN

        self.dots_on_line = 0
        self.footer_text = self.get_blank_footer_text()

    def footer(self, test_results: List[TestResult]) -> Optional[RenderableType]:
        return self.footer_text

    def get_blank_footer_text(self) -> Text:
        return Text("", end="")

    @property
    @abc.abstractmethod
    def max_dots_for_current_line(self) -> int:
        raise NotImplementedError()

    def end_of_line(self, test_index):
        self.footer_text.append(self.get_end_of_line_for_dots(test_index=test_index))
        self.console.print(self.footer_text, end="")

        self.dots_on_line = 0
        self.footer_text = self.get_blank_footer_text()

    def get_end_of_line_for_dots(
        self,
        test_index: int,
    ) -> Text:
        if TestProgressStyle.INLINE in self.progress_styles and self.num_tests > 0:
            fill = (
                self.max_dots_for_current_line - self.dots_on_line + INLINE_PROGRESS_LEN
            )
            return Text(
                f"{(test_index + 1) / self.num_tests:>{fill}.0%}\n",
                style="muted",
            )
        else:
            return Text("\n")

    def after_suite(self, test_results: List[TestResult]) -> None:
        self.end_of_line(test_index=len(test_results) - 1)


class DotsGlobal(DotsDisplayWidget):
    @property
    def max_dots_for_current_line(self) -> int:
        return self.base_max_dots_per_line

    def after_test(self, test_index: int, test_result: TestResult) -> None:
        self.footer_text.append(get_dot(test_result))

        self.dots_on_line += 1
        if self.dots_on_line == self.max_dots_for_current_line:
            self.end_of_line(test_index)


class DotsPerModule(DotsDisplayWidget):
    def __init__(self, num_tests: int, progress_styles: List[TestProgressStyle]):
        super().__init__(num_tests, progress_styles)

        self.current_path = Path("")
        self.cwd = Path.cwd()

        self._max_dots_for_current_line = self.base_max_dots_per_line

    @property
    def max_dots_for_current_line(self) -> int:
        return self._max_dots_for_current_line

    def after_test(self, test_index: int, test_result: TestResult) -> None:
        # if we are starting a new module
        if test_result.test.path != self.current_path:
            # if this isn't the first module, add the end-of-line for the previous module
            if test_index > 0:
                self.end_of_line(test_index)

            self.current_path = test_result.test.path
            rel_path = str(self.current_path.relative_to(self.cwd))

            final_slash_idx = rel_path.rfind("/")
            if final_slash_idx != -1:
                path_text = Text("", end="").join(
                    [
                        Text(rel_path[: final_slash_idx + 1], style="muted"),
                        Text(rel_path[final_slash_idx + 1 :]),
                        Text(": "),
                    ]
                )
            else:
                path_text = Text(f"{rel_path}: ", end="")

            self.footer_text.append(path_text)

            self._max_dots_for_current_line = (
                self.base_max_dots_per_line - path_text.cell_len
            )

        if self.dots_on_line == self.max_dots_for_current_line:
            self.end_of_line(test_index)

            # we are now on a blank line with no path prefix
            self._max_dots_for_current_line = self.base_max_dots_per_line

        self.footer_text.append(get_dot(test_result))
        self.dots_on_line += 1


GREEN_CHECK = Text("✔", style="pass.textonly")
RED_X = Text("✘", style="fail.textonly")


class LiveTestBar(TestResultDisplayWidget):
    def __init__(self, num_tests: int, progress_styles: List[TestProgressStyle]):
        super().__init__(num_tests, progress_styles)

        self.spinner_column = SpinnerColumn(
            style="pass.textonly",
            finished_text=GREEN_CHECK,
        )
        self.test_description_column = RenderableColumn(Text(""))

        self.progress = Progress(
            self.spinner_column,
            self.test_description_column,
            console=rich_console,
        )

        self.task = self.progress.add_task("", total=num_tests)

    def footer(self, test_results: List[TestResult]) -> Optional[RenderableType]:
        return self.progress

    def after_test(self, test_index: int, test_result: TestResult) -> None:
        self.progress.update(self.task, advance=1)
        self.test_description_column.renderable = get_test_result_line(
            test_result=test_result,
            test_index=test_index,
            num_tests=self.num_tests,
            progress_styles=self.progress_styles,
        )

        if test_result.outcome.will_fail_session:
            self.console.print(
                get_test_result_line(
                    test_result=test_result,
                    test_index=test_index,
                    num_tests=self.num_tests,
                    progress_styles=self.progress_styles,
                    extra_left_pad=2,  # account for the spinner
                )
            )

            self.spinner_column.finished_text = RED_X
            self.spinner_column.spinner.style = "fail.textonly"


class SuiteProgressBar(TestResultDisplayWidget):
    def __init__(self, num_tests: int, progress_styles: List[TestProgressStyle]):
        super().__init__(num_tests, progress_styles)

        self.spinner_column = SpinnerColumn(
            style="pass.textonly",
            finished_text=GREEN_CHECK,
        )
        self.bar_column = BarColumn(
            complete_style="pass.textonly",
            finished_style="pass.textonly",
        )

        self.progress = Progress(
            self.spinner_column,
            TimeElapsedColumn(),
            self.bar_column,
            "[progress.percentage]{task.percentage:>3.0f}%",
            "[progress.percentage][{task.completed} / {task.total}]",
            console=self.console,
        )

        self.task = self.progress.add_task("Testing...", total=num_tests)

    def footer(self, test_results: List[TestResult]) -> Optional[RenderableType]:
        return self.progress

    def after_test(self, test_index: int, test_result: TestResult) -> None:
        self.progress.update(self.task, advance=1)

        if test_result.outcome.will_fail_session:
            self.spinner_column.finished_text = RED_X
            self.spinner_column.spinner.style = "fail.textonly"
            self.bar_column.complete_style = "fail.textonly"
            self.bar_column.finished_style = "fail.textonly"

    def after_suite(self, test_results: List[TestResult]) -> None:
        self.progress = None


class TerminalResultsWriter:
    def __init__(
        self,
        console: Console,
        num_tests: int,
        progress_styles: List[TestProgressStyle],
        widget_types: Iterable[Type[TestResultDisplayWidget]],
    ):
        self.console = console
        self.widgets = [
            widgets(num_tests=num_tests, progress_styles=progress_styles)
            for widgets in widget_types
        ]
        self.live = Live(
            console=console,
            renderable=self.footer(results=[]),
        )

    def footer(self, results: List[TestResult]) -> RenderableType:
        table = Table.grid()
        table.add_column()
        for f in filter(
            None, (component.footer(results) for component in self.widgets)
        ):
            table.add_row(f)

        return table

    def run(
        self,
        test_results: Iterator[TestResult],
        fail_limit: Optional[int],
    ) -> Tuple[List[TestResult], bool]:
        """
        Execute the test suite, returning the list of test results
        and a boolean that is true if the run was cancelled and false otherwise.
        """
        num_failures = 0
        results = []
        was_cancelled = False

        self.console.print()
        with self.live as live:
            try:
                for idx, result in enumerate(test_results):
                    # We need to re-enable the Live here in case
                    # it was disabled by the breakpoint debugger hook.
                    live.start(refresh=True)

                    for component in self.widgets:
                        component.after_test(idx, result)

                    live.update(self.footer(results))

                    results.append(result)

                    if result.outcome is TestOutcome.FAIL:
                        num_failures += 1

                    if num_failures == fail_limit:
                        break
            except KeyboardInterrupt:
                was_cancelled = True
            finally:
                for component in self.widgets:
                    component.after_suite(results)

                live.update(self.footer(results), refresh=True)

                return results, was_cancelled


class TestResultWriterBase:
    runtime_output_strategies = {
        TestOutputStyle.TEST_PER_LINE: TestPerLine,
        TestOutputStyle.DOTS_GLOBAL: DotsGlobal,
        TestOutputStyle.DOTS_MODULE: DotsPerModule,
        TestOutputStyle.LIVE: LiveTestBar,
        TestOutputStyle.NONE: TestResultDisplayWidget,
    }

    def __init__(
        self,
        console: Console,
        suite: Suite,
        test_output_style: TestOutputStyle,
        progress_styles: List[TestProgressStyle],
        config_path: Optional[Path],
        show_diff_symbols: bool = False,
    ):
        self.console = console
        self.suite = suite
        self.test_output_style = test_output_style
        self.progress_styles = progress_styles
        self.config_path = config_path
        self.show_diff_symbols = show_diff_symbols
        self.terminal_size = get_terminal_size()

    def output_all_test_results(
        self,
        test_results_gen: Generator[TestResult, None, None],
        fail_limit: Optional[int] = None,
    ) -> List[TestResult]:
        if not self.suite.num_tests:
            return []

        widget_types = [self.runtime_output_strategies[self.test_output_style]]
        if TestProgressStyle.BAR in self.progress_styles:
            widget_types.append(SuiteProgressBar)

        all_results, was_cancelled = TerminalResultsWriter(
            console=self.console,
            num_tests=self.suite.num_tests_with_parameterisation,
            progress_styles=self.progress_styles,
            widget_types=widget_types,
        ).run(test_results_gen, fail_limit)

        if was_cancelled:
            self.console.print(
                "Run cancelled - results for tests that ran shown below.",
                style="info",
            )

        failed_test_results = [r for r in all_results if r.outcome == TestOutcome.FAIL]
        for failure in failed_test_results:
            self.output_why_test_failed_header(failure)
            self.output_test_failed_location(failure)
            self.output_why_test_failed(failure)
            self.output_captured_stderr(failure)
            self.output_captured_stdout(failure)
        if failed_test_results:
            self.print_divider()
        else:
            self.console.print()

        return all_results

    @staticmethod
    def print_divider() -> None:
        rich_console.print(Rule(style="muted"))

    def output_why_test_failed_header(self, test_result: TestResult):
        """
        Printed above the failing test output
        """
        raise NotImplementedError()

    def output_test_result_summary(
        self, test_results: List[TestResult], time_taken: float, duration: int
    ):
        raise NotImplementedError()

    def output_why_test_failed(self, test_result: TestResult):
        """
        Extended output shown for failing tests, may include further explanations,
        assertion error info, diffs, etc.
        """
        raise NotImplementedError()

    def output_captured_stderr(self, test_result: TestResult):
        raise NotImplementedError()

    def output_captured_stdout(self, test_result: TestResult):
        raise NotImplementedError()

    def output_test_failed_location(self, test_result: TestResult):
        raise NotImplementedError()


@dataclass
class TerminalSize:
    height: int
    width: int


def get_terminal_size() -> TerminalSize:
    for i in range(0, 3):
        try:
            cols, rows = os.get_terminal_size(i)
            return TerminalSize(height=rows, width=cols)
        except OSError:
            continue
    return TerminalSize(height=24, width=80)


class TestResultWriter(TestResultWriterBase):
    def output_why_test_failed_header(self, test_result: TestResult):
        test = test_result.test
        self.console.print(
            Padding(
                Rule(
                    title=Text(test.description, style="fail.header"),
                    style="fail.textonly",
                ),
                pad=(1, 0, 0, 0),
            ),
        )

    def output_why_test_failed(self, test_result: TestResult):
        err = test_result.error
        if isinstance(err, TestFailure):
            if err.operator in Comparison:
                self.console.print(self.get_source(err, test_result))
                self.console.print(self.get_pretty_comparison_failure(err))
        else:
            self.print_traceback(err)

    def get_source(self, err: TestFailure, test_result: TestResult) -> RenderableType:
        src_lines, line_num = inspect.getsourcelines(test_result.test.fn)
        src = Syntax(
            "".join(src_lines),
            "python",
            start_line=line_num,
            line_numbers=True,
            highlight_lines={err.error_line},
            background_color="default",
            theme="ansi_dark",
        )

        return Padding(src, (1, 0, 1, 4))

    def get_pretty_comparison_failure(self, err: TestFailure) -> RenderableType:
        if err.operator is Comparison.Equals:
            return self.get_pretty_failure_for_equals(err)
        elif err.operator in {Comparison.In, Comparison.NotIn}:
            return self.get_pretty_failure_for_in(err)
        else:
            return Text("", end="")

    def get_pretty_failure_for_equals(self, err: TestFailure) -> RenderableType:
        diff_msg = Text.assemble(
            ("LHS ", "pass.textonly"),
            ("vs ", "default"),
            ("RHS ", "fail.textonly"),
            ("shown below", "default"),
        )

        diff = Diff(
            err.lhs,
            err.rhs,
            width=self.terminal_size.width - 24,
            show_symbols=self.show_diff_symbols,
        )

        return RenderGroup(
            Padding(diff_msg, pad=(0, 0, 1, 2)),
            Padding(diff, pad=(0, 0, 1, 4)),
        )

    def get_pretty_failure_for_in(self, err: TestFailure) -> RenderableType:
        lhs_msg = Text.assemble(
            ("The ", "default"),
            ("item ", "pass.textonly"),
            *self.of_type(err.lhs),
        )
        lhs = Panel(
            Pretty(err.lhs),
            title=lhs_msg,
            title_align="left",
            border_style="pass.textonly",
            padding=1,
        )

        rhs_msg = Text.assemble(
            ("was not " if err.operator is Comparison.In else "was ", "bold default"),
            ("found in the ", "default"),
            ("container ", "fail.textonly"),
            *self.of_type(err.rhs),
        )
        rhs = Panel(
            Pretty(err.rhs),
            title=rhs_msg,
            title_align="left",
            border_style="fail.textonly",
            padding=1,
        )

        return Padding(RenderGroup(lhs, rhs), pad=(0, 0, 1, 2))

    def of_type(self, obj: object) -> Iterator[Tuple[str, str]]:
        yield "(of type ", "default"
        yield type(obj).__name__, "bold default"
        yield ")", "default"

    def print_traceback(self, err):
        trace = getattr(err, "__traceback__", "")
        if trace:
            # The first frame contains library internal code which is not
            # relevant to end users, so skip over it.
            trace = trace.tb_next
            tb = Traceback.from_exception(err.__class__, err, trace, show_locals=True)
            self.console.print(Padding(tb, pad=(0, 4, 1, 4)))
        else:
            self.console.print(str(err))

    def output_test_result_summary(
        self, test_results: List[TestResult], time_taken: float, show_slowest: int
    ):
        if show_slowest:
            self.console.print(TestTimingStatsPanel(test_results, show_slowest))

        result_table = Table.grid()
        result_table.add_column(justify="right")
        result_table.add_column()
        result_table.add_column()

        outcome_counts = self._get_outcome_counts(test_results)
        test_count = sum(outcome_counts.values())
        result_table.add_row(
            Padding(str(test_count), pad=HORIZONTAL_PAD, style="bold"),
            Padding("Tests Encountered", pad=HORIZONTAL_PAD),
            style="default",
        )
        for outcome, count in outcome_counts.items():
            if count > 0:
                result_table.add_row(
                    Padding(str(count), pad=HORIZONTAL_PAD, style="bold"),
                    Padding(outcome.display_name, pad=HORIZONTAL_PAD),
                    Padding(f"({100 * count / test_count:.1f}%)", pad=HORIZONTAL_PAD),
                    style=outcome_to_style(outcome),
                )

        exit_code = get_exit_code(test_results)
        if exit_code == ExitCode.SUCCESS:
            result_style = "pass.textonly"
        else:
            result_style = "fail.textonly"

        result_summary_panel = Panel(
            result_table,
            title="[b default]Results[/b default]",
            style="none",
            expand=False,
            border_style=result_style,
        )
        self.console.print(result_summary_panel)

        self.console.print(
            Rule(
                f"[b]{exit_code.clean_name}[/b] in [b]{time_taken:.2f}[/b] seconds",
                style=result_style,
            )
        )

    def output_captured_stderr(self, test_result: TestResult):
        if test_result.captured_stderr:
            captured_stderr_lines = test_result.captured_stderr.split("\n")
            self.console.print(Padding(Text("Captured stderr"), pad=(0, 0, 1, 2)))
            for line in captured_stderr_lines:
                self.console.print(Padding(line, pad=(0, 0, 0, 4)))
            self.console.print()

    def output_captured_stdout(self, test_result: TestResult):
        if test_result.captured_stdout:
            captured_stdout_lines = test_result.captured_stdout.split("\n")
            self.console.print(Padding(Text("Captured stdout"), pad=(0, 0, 1, 2)))
            for line in captured_stdout_lines:
                self.console.print(Padding(line, pad=(0, 0, 0, 4)))
            self.console.print()

    def output_test_failed_location(self, test_result: TestResult):
        if isinstance(test_result.error, TestFailure) or isinstance(
            test_result.error, AssertionError
        ):
            self.console.print(
                Padding(
                    Text(
                        f"Failed at {os.path.relpath(test_result.test.path, Path.cwd())}:{test_result.error.error_line}"
                    ),
                    pad=(1, 0, 0, 2),
                )
            )

    def _get_outcome_counts(
        self, test_results: List[TestResult]
    ) -> Dict[TestOutcome, int]:
        return {
            TestOutcome.PASS: len(
                [r for r in test_results if r.outcome == TestOutcome.PASS]
            ),
            TestOutcome.FAIL: len(
                [r for r in test_results if r.outcome == TestOutcome.FAIL]
            ),
            TestOutcome.SKIP: len(
                [r for r in test_results if r.outcome == TestOutcome.SKIP]
            ),
            TestOutcome.XFAIL: len(
                [r for r in test_results if r.outcome == TestOutcome.XFAIL]
            ),
            TestOutcome.XPASS: len(
                [r for r in test_results if r.outcome == TestOutcome.XPASS]
            ),
            TestOutcome.DRYRUN: len(
                [r for r in test_results if r.outcome == TestOutcome.DRYRUN]
            ),
        }


def outcome_to_style(outcome: TestOutcome) -> str:
    return {
        TestOutcome.PASS: "pass",
        TestOutcome.SKIP: "skip",
        TestOutcome.FAIL: "fail",
        TestOutcome.XFAIL: "xfail",
        TestOutcome.XPASS: "xpass",
        TestOutcome.DRYRUN: "dryrun",
    }[outcome]


def scope_to_style(scope: Scope) -> str:
    return {
        Scope.Test: "fixture.scope.test",
        Scope.Module: "fixture.scope.module",
        Scope.Global: "fixture.scope.global",
    }[scope]


def output_fixtures(
    fixtures: List[Fixture],
    tests: List[Test],
    show_scopes: bool,
    show_docstrings: bool,
    show_dependencies: bool,
    show_dependency_trees: bool,
):
    generated_tests = itertools.chain.from_iterable(
        test.get_parameterised_instances() for test in tests
    )

    fixture_to_tests = fixtures_used_directly_by_tests(generated_tests)

    fixtures_to_parents, fixtures_to_children = fixture_parents_and_children(fixtures)

    for module, fixtures in group_by(fixtures, key=lambda f: f.module_name).items():
        rich_console.print(Rule(Text(module, style="title")))

        for fixture in fixtures:
            fixture_tree = make_fixture_information_tree(
                fixture,
                used_by_tests=fixture_to_tests[fixture],
                fixtures_to_children=fixtures_to_children,
                fixtures_to_parents=fixtures_to_parents,
                show_scopes=show_scopes,
                show_docstrings=show_docstrings,
                show_dependencies=show_dependencies,
                show_dependency_trees=show_dependency_trees,
            )
            rich_console.print(fixture_tree)


def make_fixture_information_tree(
    fixture: Fixture,
    used_by_tests: Collection[Test],
    fixtures_to_children: FixtureHierarchyMapping,
    fixtures_to_parents: FixtureHierarchyMapping,
    show_scopes: bool,
    show_docstrings: bool,
    show_dependencies: bool,
    show_dependency_trees: bool,
) -> Tree:
    root = Tree(label=make_text_for_fixture(fixture, show_scope=show_scopes))

    if show_dependency_trees:
        max_depth = None
    elif show_dependencies:
        max_depth = 1
    else:
        max_depth = 0

    if show_docstrings and fixture.fn.__doc__ is not None:
        root.add(dedent(fixture.fn.__doc__).strip("\n"))

    if show_dependencies or show_dependency_trees:
        if fixtures_to_parents[fixture]:
            depends_on_node = root.add(label="[usedby]depends on fixtures")
            add_fixture_dependencies_to_tree(
                depends_on_node,
                fixture,
                fixtures_to_parents,
                show_scopes=show_scopes,
                max_depth=max_depth,
            )

        if fixtures_to_children[fixture]:
            used_by_node = root.add(label="[usedby]used by fixtures")
            add_fixture_dependencies_to_tree(
                used_by_node,
                fixture,
                fixtures_to_children,
                show_scopes=show_scopes,
                max_depth=max_depth,
            )

        if used_by_tests:
            used_by_tests_node = root.add("[usedby]used directly by tests")
            add_fixture_usages_by_tests_to_tree(used_by_tests_node, used_by_tests)

        if not (used_by_tests or fixtures_to_children[fixture]):
            root.add("[usedby]used by [fail]no tests or fixtures")

    return root


def add_fixture_dependencies_to_tree(
    parent: Tree,
    fixture: Fixture,
    fixtures_to_parents_or_children: FixtureHierarchyMapping,
    show_scopes: bool,
    max_depth: Optional[int],
    depth: int = 0,
) -> None:
    if max_depth is not None and depth >= max_depth:
        return

    this_layer = fixtures_to_parents_or_children[fixture]

    if not this_layer:
        return

    for dep in this_layer:
        node = parent.add(make_text_for_fixture(fixture=dep, show_scope=show_scopes))
        add_fixture_dependencies_to_tree(
            parent=node,
            fixture=dep,
            fixtures_to_parents_or_children=fixtures_to_parents_or_children,
            show_scopes=show_scopes,
            max_depth=max_depth,
            depth=depth + 1,
        )


def add_fixture_usages_by_tests_to_tree(node: Tree, used_by: Iterable[Test]) -> None:
    grouped_used_by = group_by(used_by, key=lambda t: t.description)
    for idx, (description, tests) in enumerate(grouped_used_by.items()):
        test = tests[0]
        loc = format_test_location(test)
        sep = f" [{len(tests)}]" if len(tests) > 1 else ""
        node.add(f"[muted]{loc}{sep}[/muted] {test.description}")


def make_text_for_fixture(fixture: Fixture, show_scope: bool) -> Text:
    text = Text()
    text.append(f"{fixture.path.name}:{fixture.line_number} ", style="dim")
    text.append(fixture.name, style="fixture.name")

    if show_scope:
        text.append(
            f" (scope: {fixture.scope.value})", style=scope_to_style(fixture.scope)
        )

    return text


def get_exit_code(results: Iterable[TestResult]) -> ExitCode:
    if not results:
        return ExitCode.NO_TESTS_FOUND

    if any(
        r.outcome == TestOutcome.FAIL or r.outcome == TestOutcome.XPASS for r in results
    ):
        exit_code = ExitCode.FAILED
    else:
        exit_code = ExitCode.SUCCESS
    return exit_code



from dataclasses import dataclass
from enum import Enum
from inspect import BoundArguments
from pathlib import Path
from typing import Callable, List, Optional, Union

from ward._errors import FixtureError

__all__ = ["Scope", "SkipMarker", "XfailMarker", "ExitCode", "CollectionMetadata"]


class Scope(Enum):
    """
    The scope of a fixture defines how long it will be cached for.

    Attributes:
        Test: A test-scoped fixture will be called each time a dependent test runs.
        Module: A module-scoped fixture will be called at most once per test module.
        Global: A global-scoped fixture will be called at most once per invocation of ``ward``.
    """

    Test = "test"
    Module = "module"
    Global = "global"

    @classmethod
    def from_str(cls, scope_str: str) -> "Scope":
        try:
            return cls[scope_str.title()]
        except (AttributeError, KeyError) as err:
            raise FixtureError(f"Invalid fixture scope: '{scope_str}'") from err


@dataclass
class Marker:
    name: str
    reason: Optional[str] = None
    when: Union[bool, Callable] = True

    @property
    def active(self):
        try:
            return self.when()
        except TypeError:
            return self.when


@dataclass
class SkipMarker(Marker):
    """
    Marker that gets attached to a test (via CollectionMetadata) to indicate it should be skipped.
    """

    name: str = "SKIP"


@dataclass
class XfailMarker(Marker):
    """
    Marker that gets attached to a test (via CollectionMetadata) to indicate that we expect it to fail.
    """

    name: str = "XFAIL"


@dataclass
class CollectionMetadata:
    """
    Attached to tests and fixtures during the collection phase for later use.
    """

    marker: Optional[Marker] = None
    description: Optional[str] = None
    tags: Optional[List[str]] = None
    is_fixture: bool = False
    scope: Scope = Scope.Test
    bound_args: Optional[BoundArguments] = None
    path: Optional[Path] = None


class ExitCode(Enum):
    """
    Enumeration of the possible exit codes that Ward can exit with.
    """

    SUCCESS = 0
    FAILED = 1
    ERROR = 2
    NO_TESTS_FOUND = 3

    @property
    def clean_name(self):
        return self.name.replace("_", " ")



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Optional, Tuple

from cucumber_tag_expressions.model import Expression

__all__ = ["Config"]


@dataclass
class Config:
    """Dataclass providing access to the user configuration that has been supplied to Ward"""

    config_path: Optional[Path]
    project_root: Optional[Path]
    path: Tuple[str]
    exclude: Tuple[str]
    search: Optional[str]
    tags: Optional[Expression]
    fail_limit: Optional[int]
    test_output_style: str
    order: str
    capture_output: bool
    show_slowest: int
    show_diff_symbols: bool
    dry_run: bool
    hook_module: Tuple[str]
    progress_style: Tuple[str]
    plugin_config: Dict[str, Dict[str, Any]]



import inspect
import types
from dataclasses import dataclass
from enum import Enum
from typing import Any, ContextManager, Generic, Optional, Type, TypeVar, cast

__all__ = [
    "raises",
    "assert_equal",
    "assert_not_equal",
    "assert_in",
    "assert_not_in",
    "assert_is",
    "assert_is_not",
    "assert_less_than",
    "assert_less_than_equal_to",
    "assert_greater_than",
    "assert_greater_than_equal_to",
]

_E = TypeVar("_E", bound=Exception)


class raises(Generic[_E], ContextManager["raises[_E]"]):
    raised: _E

    def __init__(self, expected_ex_type: Type[_E]):
        self.expected_ex_type = expected_ex_type

    def __enter__(self) -> "raises[_E]":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[types.TracebackType],
    ) -> bool:
        if exc_type is not self.expected_ex_type:
            raise AssertionError(
                f"Expected exception {self.expected_ex_type}, but {exc_type} was raised instead."
            )
        self.raised: _E = cast(_E, exc_val)
        return True


class Comparison(Enum):
    Equals = "=="
    NotEquals = "!="
    In = "in"
    NotIn = "not in"
    Is = "is"
    IsNot = "is not"
    LessThan = "<"
    LessThanEqualTo = "<="
    GreaterThan = ">"
    GreaterThanEqualTo = ">="


@dataclass
class TestFailure(Exception):
    def __init__(
        self,
        message: str,
        lhs: Any,
        rhs: Any,
        error_line: int,
        operator: Comparison,
        assert_msg: str,
    ):
        self.lhs = lhs
        self.rhs = rhs
        self.message = message
        self.error_line = error_line
        self.operator = operator
        self.assert_msg = assert_msg


def assert_equal(lhs_val: Any, rhs_val: Any, assert_msg: str) -> None:
    """
    Check whether two objects are equal. Raises a ``TestFailure`` if not.
    Args:
        lhs_val: The value on the left side of ``==``
        rhs_val: The value on the right side of ``==``
        assert_msg: The assertion message from the ``assert`` statement

    Returns: None
    Raises: TestFailure
    """
    if lhs_val != rhs_val:
        error_line_no = inspect.currentframe().f_back.f_lineno
        raise TestFailure(
            f"{lhs_val} does not equal {rhs_val}",
            lhs=lhs_val,
            rhs=rhs_val,
            error_line=error_line_no,
            operator=Comparison.Equals,
            assert_msg=assert_msg,
        )


def assert_not_equal(lhs_val: Any, rhs_val: Any, assert_msg: str) -> None:
    """
    Check whether two objects are not equal to each other. Raises a ``TestFailure`` if not.

    Args:
        lhs_val: The value on the left side of ``!=``
        rhs_val: The value on the right side of ``!=``
        assert_msg: The assertion message from the ``assert`` statement

    Returns: None
    Raises: TestFailure
    """
    if lhs_val == rhs_val:
        error_line_no = inspect.currentframe().f_back.f_lineno
        raise TestFailure(
            f"{lhs_val} does equal {rhs_val}",
            lhs=lhs_val,
            rhs=rhs_val,
            error_line=error_line_no,
            operator=Comparison.NotEquals,
            assert_msg=assert_msg,
        )


def assert_in(lhs_val: Any, rhs_val: Any, assert_msg: str) -> None:
    """
    Check if an object is contained within another via ``lhs_val in rhs_val``. Raises ``TestFailure`` if not.

    Args:
        lhs_val: The value on the left side of ``in``
        rhs_val: The value on the right side of ``in``
        assert_msg: The assertion message from the ``assert`` statement

    Returns: None
    Raises: TestFailure
    """
    if lhs_val not in rhs_val:
        error_line_no = inspect.currentframe().f_back.f_lineno
        raise TestFailure(
            f"{lhs_val} is not in {rhs_val}",
            lhs=lhs_val,
            rhs=rhs_val,
            error_line=error_line_no,
            operator=Comparison.In,
            assert_msg=assert_msg,
        )


def assert_not_in(lhs_val: Any, rhs_val: Any, assert_msg: str) -> None:
    """
    Check if an object is not contained within another via ``lhs_val not in rhs_val``.
    Raises ``TestFailure`` if lhs is contained within rhs.

    Args:
        lhs_val: The value on the left side of ``not in``
        rhs_val: The value on the right side of ``not in``
        assert_msg: The assertion message from the ``assert`` statement

    Returns: None
    Raises: TestFailure
    """
    if lhs_val in rhs_val:
        error_line_no = inspect.currentframe().f_back.f_lineno
        raise TestFailure(
            f"{lhs_val} is in {rhs_val}",
            lhs=lhs_val,
            rhs=rhs_val,
            error_line=error_line_no,
            operator=Comparison.NotIn,
            assert_msg=assert_msg,
        )


def assert_is(lhs_val: Any, rhs_val: Any, assert_msg: str) -> None:
    """
    Check the object identity via ``lhs_val is rhs_val``. Raises ``TestFailure`` if not identical.

    Args:
        lhs_val: The value on the left side of ``is``
        rhs_val: The value on the right side of ``is``
        assert_msg: The assertion message from the ``assert`` statement

    Returns: None
    Raises: TestFailure
    """
    if lhs_val is not rhs_val:
        error_line_no = inspect.currentframe().f_back.f_lineno
        raise TestFailure(
            f"{lhs_val} is not {rhs_val}",
            lhs=lhs_val,
            rhs=rhs_val,
            error_line=error_line_no,
            operator=Comparison.Is,
            assert_msg=assert_msg,
        )


def assert_is_not(lhs_val: Any, rhs_val: Any, assert_msg: str) -> None:
    """
    Check the object identity via ``lhs_val is not rhs_val``. Raises ``TestFailure`` if identical.

    Args:
        lhs_val: The value on the left side of ``is not``
        rhs_val: The value on the right side of ``is not``
        assert_msg: The assertion message from the ``assert`` statement

    Returns: None
    Raises: TestFailure
    """
    if lhs_val is rhs_val:
        error_line_no = inspect.currentframe().f_back.f_lineno
        raise TestFailure(
            f"{lhs_val} is {rhs_val}",
            lhs=lhs_val,
            rhs=rhs_val,
            error_line=error_line_no,
            operator=Comparison.IsNot,
            assert_msg=assert_msg,
        )


def assert_less_than(lhs_val: Any, rhs_val: Any, assert_msg: str) -> None:
    """
    Check lhs_val is less than the rhs_val via ``lhs_val < rhs_val``. Raises ``TestFailure`` if not.

    Args:
        lhs_val: The value on the left side of ``<``
        rhs_val: The value on the right side of ``<``
        assert_msg: The assertion message from the ``assert`` statement

    Returns: None
    Raises: TestFailure
    """
    if lhs_val >= rhs_val:
        error_line_no = inspect.currentframe().f_back.f_lineno
        raise TestFailure(
            f"{lhs_val} >= {rhs_val}",
            lhs=lhs_val,
            rhs=rhs_val,
            error_line=error_line_no,
            operator=Comparison.LessThan,
            assert_msg=assert_msg,
        )


def assert_less_than_equal_to(lhs_val: Any, rhs_val: Any, assert_msg: str) -> None:
    """
    Check lhs_val is less than or equal to the rhs_val via ``lhs_val <= rhs_val``. Raises ``TestFailure`` if not.

    Args:
        lhs_val: The value on the left side of ``<=``
        rhs_val: The value on the right side of ``<=``
        assert_msg: The assertion message from the ``assert`` statement

    Returns: None
    Raises: TestFailure
    """
    if lhs_val > rhs_val:
        error_line_no = inspect.currentframe().f_back.f_lineno
        raise TestFailure(
            f"{lhs_val} > {rhs_val}",
            lhs=lhs_val,
            rhs=rhs_val,
            error_line=error_line_no,
            operator=Comparison.LessThanEqualTo,
            assert_msg=assert_msg,
        )


def assert_greater_than(lhs_val: Any, rhs_val: Any, assert_msg: str) -> None:
    """
    Check lhs_val is greater than the rhs_val via ``lhs_val > rhs_val``. Raises ``TestFailure`` if not.

    Args:
        lhs_val: The value on the left side of ``>``
        rhs_val: The value on the right side of ``>``
        assert_msg: The assertion message from the ``assert`` statement

    Returns: None
    Raises: TestFailure
    """
    if lhs_val <= rhs_val:
        error_line_no = inspect.currentframe().f_back.f_lineno
        raise TestFailure(
            f"{lhs_val} <= {rhs_val}",
            lhs=lhs_val,
            rhs=rhs_val,
            error_line=error_line_no,
            operator=Comparison.GreaterThan,
            assert_msg=assert_msg,
        )


def assert_greater_than_equal_to(lhs_val: Any, rhs_val: Any, assert_msg: str) -> None:
    """
    Check lhs_val is greater than or equal to the rhs_val via ``lhs_val >= rhs_val``. Raises ``TestFailure`` if not.

    Args:
        lhs_val: The value on the left side of ``>=``
        rhs_val: The value on the right side of ``>=``
        assert_msg: The assertion message from the ``assert`` statement

    Returns: None
    Raises: TestFailure
    """
    if lhs_val < rhs_val:
        error_line_no = inspect.currentframe().f_back.f_lineno
        raise TestFailure(
            f"{lhs_val} < {rhs_val}",
            lhs=lhs_val,
            rhs=rhs_val,
            error_line=error_line_no,
            operator=Comparison.GreaterThanEqualTo,
            assert_msg=assert_msg,
        )



from pathlib import Path
from typing import Dict, Iterable, Optional, Union

import click
import tomli

from ward._utilities import find_project_root

_ConfigValue = Union[int, str, bool, Iterable[str]]
_ConfigDict = Dict[str, _ConfigValue]

_CONFIG_FILE = "pyproject.toml"


def _breakpoint_supported() -> bool:
    try:
        breakpoint
    except NameError:
        return False
    return True


def read_config_toml(project_root: Path, config_file: str) -> _ConfigDict:
    path = project_root / config_file
    if not path.is_file():
        return {}

    try:
        pyproject_toml = tomli.loads(path.read_text(encoding="utf-8"))
    except (tomli.TOMLDecodeError, OSError) as e:
        raise click.FileError(
            filename=config_file, hint=f"Error reading {config_file}:\n{e}"
        )

    ward_config = pyproject_toml.get("tool", {}).get("ward", {})
    ward_config = {
        k.replace("--", "").replace("-", "_"): v for k, v in ward_config.items()
    }

    return ward_config


def as_list(conf: _ConfigDict):
    if isinstance(conf, list):
        return conf
    else:
        return [conf]


def apply_multi_defaults(
    file_config: _ConfigDict,
    cli_config: _ConfigDict,
) -> _ConfigDict:
    """
    Returns all options where multiple=True that
    appeared in the config file, but weren't passed
    via the command line.
    """

    cli_paths = cli_config.get("path")
    conf_file_paths = file_config.get("path", ".")
    file_config_only = {}
    if conf_file_paths and not cli_paths:
        file_config_only["path"] = as_list(conf_file_paths)

    # TODO: Can we retrieve the tuple below programmatically?
    multiple_options = ("exclude", "hook_module")
    for param in multiple_options:
        from_cli = cli_config.get(param)
        from_conf_file = file_config.get(param, "")
        if from_conf_file and not from_cli:
            file_config_only[param] = as_list(from_conf_file)

    return file_config_only


def set_defaults_from_config(
    context: click.Context,
    param: click.Parameter,
    value: Union[str, int],
) -> Optional[Path]:
    paths_supplied_via_cli = context.params.get("path")

    search_paths = paths_supplied_via_cli
    if not search_paths:
        search_paths = (".",)

    if not context.default_map:
        context.default_map = {"path": (".",)}

    project_root = find_project_root([Path(path) for path in search_paths])
    if project_root:
        context.params["project_root"] = project_root
    else:
        context.params["project_root"] = None
        context.params["config_path"] = None
        return Path.cwd()

    file_config = read_config_toml(project_root, _CONFIG_FILE)

    if file_config:
        config_path = project_root / _CONFIG_FILE
    else:
        config_path = None

    context.params["config_path"] = config_path

    multi_defaults = apply_multi_defaults(file_config, context.params)
    file_config.update(multi_defaults)

    # Paths supplied via the CLI should be treated as relative to pwd
    # However, paths supplied via the pyproject.toml should be relative
    # to the directory that file is contained in.
    path_config_keys = ["path", "exclude"]
    for conf_key, paths in file_config.items():
        if conf_key in path_config_keys:
            relative_path_strs = []
            for path_str in paths:
                relative_path_strs.append(str((project_root / path_str)))
            file_config[conf_key] = tuple(relative_path_strs)

    context.default_map.update(file_config)

    return config_path



import pdb
import sys
from pathlib import Path
from random import shuffle
from timeit import default_timer
from typing import List, Optional, Tuple

import click
import click_completion
from click_default_group import DefaultGroup
from cucumber_tag_expressions import parse as parse_tags
from cucumber_tag_expressions.model import Expression
from rich.console import ConsoleRenderable

from ward._collect import (
    configure_path,
    filter_fixtures,
    filter_tests,
    get_info_for_modules,
    get_tests_in_modules,
    load_modules,
)
from ward._config import set_defaults_from_config
from ward._debug import init_breakpointhooks
from ward._rewrite import rewrite_assertions_in_tests
from ward._suite import Suite
from ward._terminal import (
    SessionPrelude,
    TestOutputStyle,
    TestProgressStyle,
    TestResultWriter,
    get_exit_code,
    output_fixtures,
    rich_console,
)
from ward._ward_version import __version__
from ward.config import Config
from ward.fixtures import _DEFINED_FIXTURES
from ward.hooks import plugins, register_hooks_in_modules

click_completion.init()


def _register_hooks(context: click.Context, param: click.Parameter, hook_module_names):
    register_hooks_in_modules(plugin_manager=plugins, module_names=hook_module_names)


# TODO: simplify to use invoke_without_command and ctx.forward
# once https://github.com/pallets/click/issues/430 is resolved
@click.group(
    context_settings={"max_content_width": 100},
    cls=DefaultGroup,
    default="test",
    default_if_no_args=True,
)
@click.pass_context
def run(ctx: click.Context):
    pass


config_option = click.option(
    "--config",
    type=click.Path(
        exists=False, file_okay=True, dir_okay=False, readable=True, allow_dash=False
    ),
    callback=set_defaults_from_config,
    help="Read configuration from PATH.",
    is_eager=True,
)
path_option = click.option(
    "-p",
    "--path",
    type=click.STRING,
    multiple=True,
    is_eager=True,
    help="Look for tests in PATH.",
)
exclude_option = click.option(
    "--exclude",
    type=click.STRING,
    multiple=True,
    help="Paths to ignore while searching for tests.",
)
hook_module = click.option(
    "--hook-module",
    type=click.STRING,
    callback=_register_hooks,
    multiple=True,
    help="Modules to search for hook implementations in.",
)


@run.command()
@config_option
@path_option
@exclude_option
@hook_module
@click.option(
    "--search",
    help="Search test names, bodies, descriptions and module names for the search query and only keep matching tests.",
)
@click.option(
    "--tags",
    help="Find tests matching a tag expression (e.g. 'unit and not slow').",
    metavar="EXPR",
    type=parse_tags,
)
@click.option(
    "--fail-limit",
    type=int,
    help="The maximum number of failures that are allowed to occur in a run before it is automatically cancelled.",
)
@click.option(
    "--test-output-style",
    type=click.Choice([s.value for s in TestOutputStyle], case_sensitive=False),
    default="test-per-line",
    help="The style of output for displaying individual test results during the run.",
)
@click.option(
    "--progress-style",
    type=click.Choice([s.value for s in TestProgressStyle], case_sensitive=False),
    multiple=True,
    default=["inline"],
    help="""\
    The style of progress indicator to use during the run.
    Pass multiple times to enable multiple styles.
    """,
)
@click.option(
    "--order",
    type=click.Choice(["standard", "random"], case_sensitive=False),
    default="standard",
    help="Specify the order in which tests should run.",
)
@click.option(
    "--show-diff-symbols/--hide-diff-symbols",
    default=False,
    help="""\
    If enabled, diffs will use symbols such as '?', '-', '+' and '^'
    instead of colours to highlight differences.
    """,
)
@click.option(
    "--capture-output/--no-capture-output",
    default=True,
    help="Enable or disable output capturing.",
)
@click.option(
    "--show-slowest",
    type=int,
    help="Record and display duration of n longest running tests",
    default=0,
)
@click.option(
    "--dry-run/--no-dry-run",
    help="Print all tests without executing them",
    default=False,
)
@click.version_option(version=__version__)
@click.pass_context
def test(
    ctx: click.Context,
    config: Optional[Path],
    project_root: Optional[Path],  # None if the project root cant be found
    config_path: Optional[Path],  # added by callback on '--config' option
    path: Tuple[str],
    exclude: Tuple[str],
    search: Optional[str],
    tags: Optional[Expression],
    fail_limit: Optional[int],
    test_output_style: str,
    progress_style: List[str],
    order: str,
    capture_output: bool,
    show_slowest: int,
    show_diff_symbols: bool,
    dry_run: bool,
    hook_module: Tuple[str],
):
    """Run tests."""
    config_params = ctx.params.copy()
    config_params.pop("config")

    config = Config(**config_params, plugin_config=config_params.get("plugins", {}))

    test_output_style = TestOutputStyle(test_output_style)
    progress_styles = [TestProgressStyle(ps) for ps in progress_style]

    init_breakpointhooks(pdb, sys)
    start_run = default_timer()

    print_before: Tuple[ConsoleRenderable] = plugins.hook.before_session(config=config)

    configure_path(project_root)
    paths = [Path(p) for p in path]
    mod_infos = get_info_for_modules(paths, exclude)
    modules = load_modules(mod_infos)
    unfiltered_tests = get_tests_in_modules(modules, capture_output)
    plugins.hook.preprocess_tests(config=config, collected_tests=unfiltered_tests)
    filtered_tests = filter_tests(unfiltered_tests, query=search, tag_expr=tags)
    if config.order == "random":
        shuffle(filtered_tests)

    tests = rewrite_assertions_in_tests(filtered_tests)

    time_to_collect_secs = default_timer() - start_run

    suite = Suite(tests=tests)
    test_results = suite.generate_test_runs(dry_run=dry_run)
    rich_console.print(
        SessionPrelude(
            time_to_collect_secs=time_to_collect_secs,
            num_tests_collected=suite.num_tests_with_parameterisation,
            num_fixtures_collected=len(_DEFINED_FIXTURES),
            config_path=config_path,
        )
    )
    writer = TestResultWriter(
        console=rich_console,
        suite=suite,
        test_output_style=test_output_style,
        progress_styles=progress_styles,
        config_path=config_path,
        show_diff_symbols=show_diff_symbols,
    )
    for renderable in print_before:
        rich_console.print(renderable)
    test_results = writer.output_all_test_results(test_results, fail_limit=fail_limit)
    exit_code = get_exit_code(test_results)
    time_taken = default_timer() - start_run

    render_afters: Tuple[ConsoleRenderable] = plugins.hook.after_session(
        config=config, test_results=test_results, status_code=exit_code
    )
    for renderable in render_afters:
        rich_console.print(renderable)

    writer.output_test_result_summary(test_results, time_taken, show_slowest)
    sys.exit(exit_code.value)


@run.command()
@config_option
@path_option
@exclude_option
@click.option(
    "-f",
    "--fixture-path",
    help="Only display fixtures defined in or below the given paths.",
    multiple=True,
    type=Path,
)
@click.option(
    "--search",
    help="Search fixtures names, bodies, and module names for the search query and only keep matching fixtures.",
)
@click.option(
    "--show-scopes/--no-show-scopes",
    help="Display each fixture's scope.",
    default=True,
)
@click.option(
    "--show-docstrings/--no-show-docstrings",
    help="Display each fixture's docstring.",
    default=False,
)
@click.option(
    "--show-dependencies/--no-show-dependencies",
    help="Displays direct dependencies between fixtures and tests."
    "Use --show-dependency-trees to show detailed dependency information.",
    default=False,
)
@click.option(
    "--show-dependency-trees/--no-show-dependency-trees",
    help="Display the entire dependency tree for each fixture.",
    default=False,
)
@click.option(
    "--full/--no-full",
    help="Display all available information on each fixture.",
    default=False,
)
@click.pass_context
def fixtures(
    ctx: click.Context,
    config: str,
    project_root: Optional[Path],
    config_path: Optional[Path],
    path: Tuple[str],
    exclude: Tuple[str],
    fixture_path: Tuple[Path],
    search: Optional[str],
    show_scopes: bool,
    show_docstrings: bool,
    show_dependencies: bool,
    show_dependency_trees: bool,
    full: bool,
):
    """Show information on fixtures."""
    configure_path(project_root)
    paths = [Path(p) for p in path]
    mod_infos = get_info_for_modules(paths, exclude)
    modules = list(load_modules(mod_infos))
    tests = list(get_tests_in_modules(modules, capture_output=True))

    filtered_fixtures = list(
        filter_fixtures(_DEFINED_FIXTURES, query=search, paths=fixture_path)
    )

    output_fixtures(
        fixtures=filtered_fixtures,
        tests=tests,
        show_scopes=show_scopes or full,
        show_docstrings=show_docstrings or full,
        show_dependencies=show_dependencies or full,
        show_dependency_trees=show_dependency_trees or full,
    )


@run.command()
@click.pass_context
def completions(ctx: click.Context):
    shell, path = click_completion.core.install()
    click.echo(f"{shell} completion installed in {path}")
    ctx.exit(0)



import importlib
import inspect
import io
import os
import sys
from typing import Any

import click

from ward._config import _breakpoint_supported
from ward._terminal import rich_console

original_stdout = sys.stdout


def init_breakpointhooks(pdb_module, sys_module) -> None:
    # breakpoint() is Python 3.7+
    if _breakpoint_supported():
        sys_module.breakpointhook = _breakpointhook
    pdb_module.set_trace = _breakpointhook


def _breakpointhook(*args, **kwargs):
    hookname = os.getenv("PYTHONBREAKPOINT")
    if hookname is None or len(hookname) == 0:
        hookname = "pdb.set_trace"
        kwargs.setdefault("frame", inspect.currentframe().f_back)
    elif hookname == "0":
        return None

    hook = _get_debugger_hook(hookname)
    context = click.get_current_context()
    capture_enabled = context.params.get("capture_output")
    capture_active = isinstance(sys.stdout, io.StringIO)

    # Stop an active Live.
    # It is the responsibility of the test executor
    # to re-enable the Live.
    live = rich_console._live
    if live is not None:
        live.refresh()
        live.stop()

    if capture_enabled and capture_active:
        sys.stdout = original_stdout
        rich_console.print(
            f"Entering debugger [bold]{hookname}[/] - output capturing disabled.",
            style="info",
        )

    return hook(*args, **kwargs)


def _get_debugger_hook(hookname: str):
    modname, dot, funcname = hookname.rpartition(".")
    if dot == "":
        modname = "builtins"
    module: Any = importlib.import_module(modname)
    if hookname == "pdb.set_trace":
        set_trace = module.Pdb(stdout=original_stdout, skip=["ward*"]).set_trace
        hook = set_trace
    else:
        hook = getattr(module, funcname)
    return hook


__breakpointhook__ = _breakpointhook



import importlib
from typing import Iterable, List, Optional

import pluggy
from rich.console import ConsoleRenderable

from ward.config import Config
from ward.models import ExitCode
from ward.testing import Test, TestResult

PROJECT_NAME = "ward"

spec = pluggy.HookspecMarker(PROJECT_NAME)
hook = pluggy.HookimplMarker(PROJECT_NAME)


def register_hooks_in_modules(
    plugin_manager: pluggy.PluginManager, module_names: Iterable[str]
):
    for module_name in module_names:
        _register_hooks_in_module(plugin_manager, module_name)


def _register_hooks_in_module(plugin_manager: pluggy.PluginManager, module_name: str):
    module = importlib.import_module(module_name)
    plugin_manager.register(module)


class SessionHooks:
    @spec
    def before_session(self, config: Config) -> Optional[ConsoleRenderable]:
        """
        Hook which is called immediately at the start of a test session (before tests are collected).

        You can implement this hook to run some setup code.

        This hook has no default implementation. If you implement it, you will not be
        overriding any existing functionality.

        Examples of how you could use this hook:

        * Printing some information to the terminal about your plugin.
        * Creating a file on disk which you'll write to in other hooks.

        If you return a
        `rich.console.ConsoleRenderable<https://rich.readthedocs.io/en/latest/protocol.html#console-render>`_
        from this function, it will be rendered to the terminal.
        """

    @spec
    def after_session(
        self,
        config: Config,
        test_results: List[TestResult],
        status_code: ExitCode,
    ) -> Optional[ConsoleRenderable]:
        """
        Hook that runs right before a test session ends (just before the result summary is printed to the terminal).

        This hook has no default implementation. If you implement it, you will not be
        overriding any existing functionality.

        Examples of how you could use this hook:

        * Writing additional information to the terminal after the test session finishes.
        * Writing a file (e.g. an HTML report) to disk containing the results from the test session.
        * Sending a file containing the results off somewhere for storage/processing.

        If you return a
        `rich.console.ConsoleRenderable <https://rich.readthedocs.io/en/latest/protocol.html#console-render>`_
        from this function, it will be rendered to the terminal.
        """

    @spec
    def preprocess_tests(self, config: Config, collected_tests: List[Test]):
        """
        Called before tests are filtered out by ``--search``, ``--tags``, etc, and before assertion rewriting.

        Allows you to transform or filter out tests (by modifying the ``collected_tests`` list in place).

        This hook has no default implementation. You can implement it without overriding existing functionality.

        Examples of how you could use this hook:

        * Filter out tests
        * Reorder tests
        * Generate statistics about tests pre-run
        * Attach some data to each test to use later
        """


plugins = pluggy.PluginManager(PROJECT_NAME)
plugins.add_hookspecs(SessionHooks)
plugins.load_setuptools_entrypoints(PROJECT_NAME)



import asyncio
import inspect
from contextlib import suppress
from dataclasses import dataclass
from functools import partial, wraps
from pathlib import Path
from typing import Any, AsyncGenerator, Callable, Generator, List, Union

from ward.models import CollectionMetadata, Scope

__all__ = ["fixture", "using", "Fixture"]


@dataclass
class Fixture:
    """
    Represents a piece of data that will be used in a test.

    Attributes:
        fn: The Python function object corresponding to this fixture.
        gen: The generator, if applicable to this fixture.
        resolved_val: The value returned by calling the fixture function (fn).
    """

    fn: Callable
    gen: Union[Generator, AsyncGenerator, None] = None
    resolved_val: Any = None

    def __hash__(self):
        return hash(self._id)

    def __eq__(self, other):
        return self._id == other._id

    @property
    def _id(self):
        return self.__class__, self.name, self.path, self.line_number

    @property
    def key(self) -> str:
        """
        A unique key used to identify fixture in the fixture cache. A string of the form '{path}::{name}'
        """
        path = self.path
        name = self.name
        return f"{path}::{name}"

    @property
    def scope(self) -> Scope:
        return getattr(self.fn, "ward_meta").scope

    @property
    def name(self):
        """
        The name of the fixture function.
        """
        return self.fn.__name__

    @property
    def path(self):
        """
        The pathlib.Path of the module the fixture is defined in.
        """
        return self.fn.ward_meta.path

    @property
    def module_name(self):
        """
        The name of the module the fixture is defined in.
        """
        return self.fn.__module__

    @property
    def qualified_name(self) -> str:
        name = self.name or ""
        return f"{self.module_name}.{name}"

    @property
    def line_number(self) -> int:
        """
        The line number that the fixture is defined on.
        """
        return inspect.getsourcelines(self.fn)[1]

    @property
    def is_generator_fixture(self):
        """
        True if the fixture is a generator function (and thus contains teardown code).
        """
        return inspect.isgeneratorfunction(inspect.unwrap(self.fn))

    @property
    def is_async_generator_fixture(self):
        """
        True if this fixture is an async generator.
        """
        return inspect.isasyncgenfunction(inspect.unwrap(self.fn))

    @property
    def is_coroutine_fixture(self):
        """
        True if the fixture is defined with 'async def'.
        """
        return inspect.iscoroutinefunction(inspect.unwrap(self.fn))

    def deps(self):
        """
        The dependencies of the fixture.
        """
        return inspect.signature(self.fn).parameters

    def parents(self) -> List["Fixture"]:
        """
        Return the parent fixtures of this fixture, as a list of Fixtures.
        """
        return [Fixture(par.default) for par in self.deps().values()]

    def teardown(self):
        """
        Tears down the fixture by calling `next` or `__anext__()`.
        """
        # Suppress because we can't know whether there's more code
        # to execute below the yield.
        with suppress(RuntimeError, StopIteration, StopAsyncIteration):
            if self.is_generator_fixture and self.gen:
                next(self.gen)
            elif self.is_async_generator_fixture and self.gen:
                awaitable = self.gen.__anext__()
                asyncio.get_event_loop().run_until_complete(awaitable)


def fixture(func=None, *, scope: Union[Scope, str] = Scope.Test):
    """
    Decorator which will cause the wrapped function to be collected and treated as a fixture.

    Args:
        func: The wrapped function which should yield or return some data required to execute a test.
        scope: The scope of a fixture determines how long it can be cached for (and therefore how frequently
            the fixture should be regenerated).
    """
    if not isinstance(scope, Scope):
        scope = Scope.from_str(scope)

    if func is None:
        return partial(fixture, scope=scope)

    # By setting is_fixture = True, the framework will know
    # that if this fixture is provided as a default arg, it
    # is responsible for resolving the value.
    path = Path(inspect.getfile(func)).absolute()
    if hasattr(func, "ward_meta"):
        func.ward_meta.is_fixture = True
        func.ward_meta.path = path
    else:
        func.ward_meta = CollectionMetadata(is_fixture=True, scope=scope, path=path)

    _DEFINED_FIXTURES.append(Fixture(func))

    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def using(*using_args, **using_kwargs):
    """
    An alternative to the default param method of injecting fixtures into tests. Allows you to avoid using
    keyword arguments in your test definitions.
    """

    def decorator_using(func):
        signature = inspect.signature(func)
        bound_args = signature.bind_partial(*using_args, **using_kwargs)
        if hasattr(func, "ward_meta"):
            func.ward_meta.bound_args = bound_args
        else:
            func.ward_meta = CollectionMetadata(bound_args=bound_args)

        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    return decorator_using


_DEFINED_FIXTURES: List[Fixture] = []



"""A modern Python 3 test framework with a focus on productivity and readability."""
from ._ward_version import __version__
from .expect import raises
from .fixtures import fixture, using
from .models import Scope
from .testing import each, skip, test, xfail

__all__ = [
    "__version__",
    "raises",
    "fixture",
    "using",
    "Scope",
    "each",
    "skip",
    "test",
    "xfail",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ward']

package_data = \
{'': ['*']}

install_requires = \
['click-completion>=0.5.2,<0.6.0',
 'click-default-group>=1.2.2,<2.0.0',
 'click>=8.0.0,<9.0.0',
 'cucumber-tag-expressions>=2.0.0,<3.0.0',
 'pluggy>=0.13.1,<0.14.0',
 'pprintpp>=0.4.0,<0.5.0',
 'rich>=10.0.0,<11.0.0',
 'tomli>=1.0.0,<2.0.0']

extras_require = \
{':python_version == "3.6"': ['dataclasses>=0.7,<0.8']}

entry_points = \
{'console_scripts': ['ward = ward._run:run']}

setup_kwargs = {
    'name': 'ward',
    'version': '0.61.0b0',
    'description': 'A modern Python testing framework',
    'long_description': '<img src="https://user-images.githubusercontent.com/5740731/119056107-085c6900-b9c2-11eb-9699-f54ef4945623.png" width="350px">\n\n[![Codecov](https://codecov.io/gh/darrenburns/ward/branch/master/graph/badge.svg)](https://codecov.io/gh/darrenburns/ward)\n[![Documentation Status](https://readthedocs.org/projects/ward/badge/?version=latest)](https://ward.readthedocs.io/en/latest/?badge=latest)\n[![PyPI version](https://badge.fury.io/py/ward.svg)](https://badge.fury.io/py/ward)\n\n<hr>\n\n_Ward_ is a Python testing framework with a focus on productivity and readability. It gives you the tools you need to write **well-documented** and **scalable** tests.\n\n<img alt="Ward typical test output example" src="https://user-images.githubusercontent.com/5740731/118399779-a795ff00-b656-11eb-8fca-4ceb03151f3e.png">\n\n## Features\n\nSee the full set of features in the [**documentation**](https://ward.readthedocs.io).\n\n**Descriptive test names:** describe what your tests do using strings, not function names.\n```python\n@test("simple addition")  # you can use markdown in these descriptions!\ndef _():\n    assert 1 + 2 == 3  # you can use plain assert statements!\n```\n\n**Modular test dependencies:** manage test setup/teardown code using fixtures that rely on Python\'s import system, not\nname matching.\n```python\n@fixture\ndef user():\n    return User(name="darren")\n\n\n@test("the user is called darren")\ndef _(u=user):\n    assert u.name == "darren"\n```\n\n**Support for asyncio**: define your tests and fixtures with `async def` and call asynchronous code within them.\n\n```python\n@fixture\nasync def user():\n    u = await create_user()\n    return await u.login()\n\n\n@test("the logged in user has a last session date")\nasync def _(user=user):\n    last_session = await get_last_session_date(user.id)\n    assert is_recent(last_session, get_last_session_date)\n```\n\n**Powerful test selection:** limit your test run not only by matching test names/descriptions, but also on the code\ncontained in the body of the test.\n```\nward --search "Database.get_all_users"\n```\nOr use tag expressions for more powerful filtering.\n```\nward --tags "(unit or integration) and not slow"\n```\n\n**Parameterised testing:** write a test once, and run it multiple times with different inputs by writing it in a loop.\n```python\nfor lhs, rhs, res in [\n    (1, 1, 2),\n    (2, 3, 5),\n]:\n\n    @test("simple addition")\n    def _(left=lhs, right=rhs, result=res):\n        assert left + right == result\n```\n\n**Cross platform:** Tested on Mac OS, Linux, and Windows.\n\n**Speedy:** Ward\'s suite of ~320 tests run in less than half a second on my machine.\n\n**Zero config:** Sensible defaults mean running `ward` with no arguments is enough to get started. Can be configured using `pyproject.toml` or the command line if required.\n\n**Extendable:** Ward has a plugin system built with pluggy, the same framework used by pytest.\n\n**Colourful, human readable output:** quickly pinpoint and fix issues with detailed output for failing tests.\n\n<img alt="Ward failing test output example" src="https://user-images.githubusercontent.com/5740731/120125898-5dfaf780-c1b2-11eb-9acd-b9cd0ff24110.png">\n\n## Getting Started\n\nHave a look at the [**documentation**](https://ward.readthedocs.io)!\n\n## How to Contribute\n\nContributions are very welcome and encouraged!\n\nSee the [contributing guide](.github/CONTRIBUTING.md) for information on how you can take part in the development of Ward.\n',
    'author': 'Darren Burns',
    'author_email': 'darrenb900@gmail.com',
    'maintainer': 'Darren Burns',
    'maintainer_email': 'darrenb900@gmail.com',
    'url': 'https://wardpy.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



import collections
import inspect
from pathlib import Path
from typing import Any, Callable, Dict, Hashable, Iterable, List, Optional, TypeVar


def truncate(s: str, num_chars: int) -> str:
    suffix = "..." if len(s) > num_chars else ""
    return s[: num_chars - len(suffix)] + suffix


def find_project_root(paths: Iterable[Path]) -> Optional[Path]:
    if not paths:
        return None

    common_base = min(path.resolve() for path in paths)
    if common_base.is_dir():
        common_base /= "child-of-base"

    # Check this common base and all of its parents for files
    # indicating the project root
    for directory in common_base.parents:
        if (directory / "pyproject.toml").is_file():
            return directory
        if (directory / ".git").exists():
            return directory
        if (directory / ".hg").is_dir():
            return directory

    return None


def get_absolute_path(object: Any) -> Path:
    return Path(inspect.getfile(object)).absolute()


T = TypeVar("T")
H = TypeVar("H", bound=Hashable)


def group_by(items: Iterable[T], key: Callable[[T], H]) -> Dict[H, List[T]]:
    groups = collections.defaultdict(list)
    for item in items:
        groups[key(item)].append(item)
    return dict(groups)



import ast
import inspect
import textwrap
import types
from typing import Iterable, List

from ward.expect import (
    assert_equal,
    assert_greater_than,
    assert_greater_than_equal_to,
    assert_in,
    assert_is,
    assert_is_not,
    assert_less_than,
    assert_less_than_equal_to,
    assert_not_equal,
    assert_not_in,
)
from ward.testing import Test

assert_func_namespace = {
    assert_equal.__name__: assert_equal,
    assert_not_equal.__name__: assert_not_equal,
    assert_in.__name__: assert_in,
    assert_not_in.__name__: assert_not_in,
    assert_is.__name__: assert_is,
    assert_is_not.__name__: assert_is_not,
    assert_less_than_equal_to.__name__: assert_less_than_equal_to,
    assert_less_than.__name__: assert_less_than,
    assert_greater_than_equal_to.__name__: assert_greater_than_equal_to,
    assert_greater_than.__name__: assert_greater_than,
}


def get_assertion_msg(node: ast.expr) -> str:
    if node.msg and isinstance(node.msg, ast.Str):
        msg = node.msg.s
    else:
        msg = ""
    return msg


def make_call_node(node: ast.expr, func_name: str) -> ast.Expr:
    msg = get_assertion_msg(node)
    call = ast.Call(
        func=ast.Name(id=func_name, ctx=ast.Load()),
        args=[node.test.left, node.test.comparators[0], ast.Str(s=msg)],
        keywords=[],
    )
    new_node = ast.Expr(value=call)
    ast.copy_location(new_node, node)
    ast.fix_missing_locations(new_node)
    return new_node


def is_binary_comparison(node: ast.expr) -> bool:
    return isinstance(node.test, ast.Compare) and len(node.test.ops) == 1


def is_comparison_type(node: ast.expr, type) -> bool:
    return isinstance(node.test.ops[0], type)


class RewriteAssert(ast.NodeTransformer):
    def visit_Assert(self, node):  # noqa: C901 - no chance to reduce complexity
        if is_binary_comparison(node):
            if is_comparison_type(node, ast.Eq):
                return make_call_node(node, assert_equal.__name__)
            elif is_comparison_type(node, ast.NotEq):
                return make_call_node(node, assert_not_equal.__name__)
            elif is_comparison_type(node, ast.In):
                return make_call_node(node, assert_in.__name__)
            elif is_comparison_type(node, ast.NotIn):
                return make_call_node(node, assert_not_in.__name__)
            elif is_comparison_type(node, ast.Is):
                return make_call_node(node, assert_is.__name__)
            elif is_comparison_type(node, ast.IsNot):
                return make_call_node(node, assert_is_not.__name__)
            elif is_comparison_type(node, ast.Lt):
                return make_call_node(node, assert_less_than.__name__)
            elif is_comparison_type(node, ast.LtE):
                return make_call_node(node, assert_less_than_equal_to.__name__)
            elif is_comparison_type(node, ast.Gt):
                return make_call_node(node, assert_greater_than.__name__)
            elif is_comparison_type(node, ast.GtE):
                return make_call_node(node, assert_greater_than_equal_to.__name__)

        return node


def rewrite_assertions_in_tests(tests: Iterable[Test]) -> List[Test]:
    return [rewrite_assertion(test) for test in tests]


def rewrite_assertion(test: Test) -> Test:
    # Get the old code and code object
    code_lines, line_no = inspect.getsourcelines(test.fn)

    code = "".join(code_lines)
    indents = textwrap._leading_whitespace_re.findall(code)
    col_offset = len(indents[0]) if len(indents) > 0 else 0
    code = textwrap.dedent(code)
    code_obj = test.fn.__code__

    # Rewrite the AST of the code
    tree = ast.parse(code)
    ast.increment_lineno(tree, line_no - 1)

    new_tree = RewriteAssert().visit(tree)

    # We dedented the code so that it was a valid tree, now re-apply the indent
    for child in ast.walk(new_tree):
        if hasattr(child, "col_offset"):
            child.col_offset = getattr(child, "col_offset", 0) + col_offset

    # Reconstruct the test function
    new_mod_code_obj = compile(new_tree, code_obj.co_filename, "exec")

    # TODO: This probably isn't correct for nested closures
    clo_glob = {}
    if test.fn.__closure__:
        clo_glob = test.fn.__closure__[0].cell_contents.__globals__

    # Look through the new module,
    # find the code object with the same name as the original code object,
    # and build a new function with the injected assert functions added to the global namespace.
    # Filtering on the code object name prevents finding other kinds of code objects,
    # like lambdas stored directly in test function arguments.
    for const in new_mod_code_obj.co_consts:
        if isinstance(const, types.CodeType) and const.co_name == code_obj.co_name:
            new_test_func = types.FunctionType(
                const,
                {**assert_func_namespace, **test.fn.__globals__, **clo_glob},
                test.fn.__name__,
                test.fn.__defaults__,
            )
            new_test_func.ward_meta = test.fn.ward_meta
            return Test(
                **{k: vars(test)[k] for k in vars(test) if k != "fn"},
                fn=new_test_func,
            )

    return test



from contextlib import suppress
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Collection, Dict, Iterable, Mapping, Tuple, Union

from ward.fixtures import Fixture
from ward.models import Scope

FixtureKey = str
TestId = str
ScopeKey = Union[TestId, Path, Scope]
ScopeCache = Dict[Scope, Dict[ScopeKey, Dict[FixtureKey, Fixture]]]


def _scope_cache_factory():
    return {scope: {} for scope in Scope}


@dataclass
class FixtureCache:
    """
    A collection of caches, each storing data for a different scope.

    When a fixture is resolved, it is stored in the appropriate cache given
    the scope of the fixture.

    A lookup into this cache is a 3 stage process:

    Scope -> ScopeKey -> FixtureKey

    The first 2 lookups (Scope and ScopeKey) let us determine:
        e.g. has a test-scoped fixture been cached for the current test?
        e.g. has a module-scoped fixture been cached for the current test module?

    The final lookup lets us retrieve the actual fixture given a fixture key.
    """

    _scope_cache: ScopeCache = field(default_factory=_scope_cache_factory)

    def _get_subcache(self, scope: Scope) -> Dict[str, Any]:
        return self._scope_cache[scope]

    def get_fixtures_at_scope(
        self, scope: Scope, scope_key: ScopeKey
    ) -> Dict[FixtureKey, Fixture]:
        subcache = self._get_subcache(scope)
        if scope_key not in subcache:
            subcache[scope_key] = {}
        return subcache.get(scope_key)

    def cache_fixture(self, fixture: Fixture, scope_key: ScopeKey):
        """
        Cache a fixture at the appropriate scope for the given test.
        """
        fixtures = self.get_fixtures_at_scope(fixture.scope, scope_key)
        fixtures[fixture.key] = fixture

    def teardown_fixtures_for_scope(self, scope: Scope, scope_key: ScopeKey):
        fixture_dict = self.get_fixtures_at_scope(scope, scope_key)
        fixtures = list(fixture_dict.values())
        for fixture in fixtures:
            with suppress(RuntimeError, StopIteration):
                fixture.teardown()
            del fixture_dict[fixture.key]

    def teardown_global_fixtures(self):
        self.teardown_fixtures_for_scope(Scope.Global, Scope.Global)

    def contains(self, fixture: Fixture, scope: Scope, scope_key: ScopeKey) -> bool:
        fixtures = self.get_fixtures_at_scope(scope, scope_key)
        return fixture.key in fixtures

    def get(
        self, fixture_key: FixtureKey, scope: Scope, scope_key: ScopeKey
    ) -> Fixture:
        fixtures = self.get_fixtures_at_scope(scope, scope_key)
        return fixtures.get(fixture_key)


def is_fixture(obj: Any) -> bool:
    """
    Returns True if and only if the object is a fixture function
    (it would be False for a Fixture instance,
    but True for the underlying function inside it).
    """
    return hasattr(obj, "ward_meta") and obj.ward_meta.is_fixture


FixtureHierarchyMapping = Mapping[Fixture, Collection[Fixture]]


def fixture_parents_and_children(
    fixtures: Iterable[Fixture],
) -> Tuple[FixtureHierarchyMapping, FixtureHierarchyMapping]:
    """
    Given an iterable of Fixtures, produce two dictionaries:
    the first maps each fixture to its parents (the fixtures it depends on);
    the second maps each fixture to its children (the fixtures that depend on it).
    """
    fixtures_to_parents = {fixture: fixture.parents() for fixture in fixtures}

    # not a defaultdict, because we want to have empty entries if no parents when we return
    fixtures_to_children = {fixture: [] for fixture in fixtures}
    for fixture, parents in fixtures_to_parents.items():
        for parent in parents:
            fixtures_to_children[parent].append(fixture)

    return fixtures_to_parents, fixtures_to_children
